"""
    Module containing methods that convert networks into various formats
"""

import pathlib

import pandas as pd

from ..main import Network


def json_to_elist(in_file: pathlib.Path, out_file: pathlib.Path) -> None:
    """
    Convert Network file from json to elist format
    Note that only the edge attributes can be converted

    Parameters
    ----------
    in_file : pathlib.Path
        The path to the json formatted network file
    out_file : pathlib.Path
        The path to the elist formatted network file
    """
    network = Network.load_json(in_file)
    df = pd.DataFrame.from_dict(network.links)
    cols = list(df.columns)
    cols.remove("source")
    cols.remove("target")
    df = df[["source", "target"] + cols]
    df.to_csv(out_file, index=False)


CONVERTERS = {("json", "elist"): json_to_elist}



#!/usr/bin/env python3

import os

from biom import load_table
from biom.util import biom_open
import pandas as pd


HEADERS = ["Kingdom", "Phylum", "Class", "Order", "Family", "Genus", "Species"]


def tax_splitter(full_tax):
    tax_list = [tax.strip().split("__")[-1] for tax in full_tax.Taxon.split(";")]
    tax_list = [
        t.replace("'", "").replace("=", "").replace("[", "").replace("]", "")
        for t in tax_list
    ]
    tax_list.extend([""] * (len(HEADERS) - len(tax_list)))
    return pd.Series(data=tax_list, index=HEADERS)


def main(otu_table_file, tax_assignment_file, sample_metadata_file):
    otu_table = load_table(otu_table_file)
    if os.path.splitext(sample_metadata_file)[1] == "csv":
        sample_metadata = pd.read_csv(sample_metadata_file, index_col=0, comment="#")
    else:
        sample_metadata = pd.read_table(sample_metadata_file, index_col=0, comment="#")
    tax_assignment = pd.read_table(tax_assignment_file, index_col=0)
    obs_metadata = tax_assignment[["Taxon"]].apply(tax_splitter, axis=1)
    for index in otu_table.ids("observation"):
        if index not in obs_metadata.index:
            obs_metadata.loc[index] = [""] * len(HEADERS)
    otu_table.del_metadata(axis="observation")
    otu_table.add_metadata(obs_metadata.to_dict(orient="index"), axis="observation")
    otu_table.del_metadata(axis="sample")
    otu_table.add_metadata(sample_metadata.to_dict(orient="index"), axis="sample")
    with biom_open("otu_table_wtax.biom", "w") as fid:
        otu_table.to_hdf5(fid, "Constructed using the micone pipeline")


if __name__ == "__main__":
    OTU_TABLE = "${otu_table_file}"
    TAX_ASSIGNMENT = "${tax_assignment}"
    SAMPLE_METADATA = "${sample_metadata_file}"
    main(OTU_TABLE, TAX_ASSIGNMENT, SAMPLE_METADATA)



#!/usr/bin/env python3

import csv
from functools import partial
from warnings import warn


def get_score(baseid: int, stat_dict: dict) -> str:
    """
    Calculate whether a given base id is 'good', 'warning' or 'failure'
    A warning will be issued if the lower quartile for any base is less than 10,
    or if the median for any base is less than 25.
    A failure will be issued if the lower quartile for any base is less than 5,
    or if the median for any base is less than 20.

    Parameters
    ---------
    baseid : int
        Base position currently under consideration for scoring
    stat_dict : Dict[str, List[float]]
        A dictionary containing scores for each of the seven statistics

    Returns
    ------
    str
        A string that is either 'good', 'warning' or 'failure'
    good = lambda s25: s25 > 28
    """
    warning = lambda s25, s50: (s25 < 10) or (s50 < 25)
    failure = lambda s25, s50: (s25 < 5) or (s50 < 20)
    stat25, stat50 = stat_dict["25%"][baseid], stat_dict["50%"][baseid]
    if failure(stat25, stat50):
        score = "failure"
    elif warning(stat25, stat50):
        score = "warning"
    else:
        score = "good"
    return score


def trimmer(
    scores: list,
    counts: list,
    direction: int = 0,
    threshold: int = 3,
    frac_retain: float = 0.9,
) -> int:
    """
    Trimmer takes scores and direction as input and returns the position where
    the sequence has to be trimmed in that direction

    Parameters
    ---------
    scores : list
        List of scores for each baseid. Scores are 'good', 'warning' or 'failure'
    counts : list
        Number of sequences used to calculate scores for each baseid
    direction : {0, 1}, optional
        The direction for trimming. 0 is left, 1 is right
        Default value is 0
    threshold : int, optional
        The minimum number of good bases that need to appear in sequence to stop trimming
        Default value is 5
    frac_retain : float, optional
        The minimum fraction of sequences that need to be retained after trimming

    Returns
    ------
    int
        The position where the sequence is to be trimmed in the given direction
    """
    ord_scores = scores if direction == 0 else scores[::-1]
    nseqs = counts[0]
    count_fracs = [c / nseqs for c in counts]
    try:
        retain_fail_pos = next(
            ind for ind, c in enumerate(count_fracs) if c < frac_retain
        )
    except StopIteration:
        retain_fail_pos = len(scores)
    stop_flag = 0
    pos = len(scores)
    for i, score in enumerate(ord_scores):
        if stop_flag <= -threshold:
            stop_flag = 0
        if stop_flag >= threshold:
            pos = i - threshold
            break
        elif score == "good":
            stop_flag += 1
        elif score == "warning":
            stop_flag += 0
        elif score == "failure":
            stop_flag -= 1
    direct_pos = pos if direction == 0 else len(scores) - pos
    return direct_pos if direction == 0 else min(direct_pos, retain_fail_pos)


def analyze_quality(quality_file: str) -> list:
    """
    Takes in quality_file location as input and gives out the scores for each positon

    Parameters
    ---------
    quality_file : str
        The location of the quality file

    Returns
    ------
    list
        List that contains the scores 'good', 'warning', 'failure' for each position
    """
    with open(quality_file, "r") as fid:
        header, count, *stats = csv.reader(fid, delimiter="\\t")
    stat_dict = dict((stat[0], list(map(float, stat[1:]))) for stat in stats)
    if len(set(count[1:])) > 1:
        warn("Sequences are of different lengths")
    base_ids = list(map(int, header[1:]))
    scorer_fun = partial(get_score, stat_dict=stat_dict)
    scores = list(map(scorer_fun, base_ids))
    counts = list(map(float, count[1:]))
    return scores, counts


def quality_analysis(forward_quality: str) -> None:
    """
    A program to identify the how many bases needs to be trimmed off from the
    start and end of our sequences

    Parameters
    ----------
    forward_quality : str
        The file containing the seven number quality information for each base of forward read
        (2%, 9%, 25%, 50%, 75%, 91%, 98%)
    """
    with open("trim.txt", "w") as fid:
        scores, counts = analyze_quality(forward_quality)
        left_trim = trimmer(scores, counts)
        length_trim = trimmer(scores, counts, direction=1)
        # fid.write("--cut={}\\n".format(left_trim))
        # fid.write("--length={}\\n".format(length_trim))
        # fid.write("--minimum-length={}\\n".format(length_trim))
        # fid.write("--maximum-length={}".format(length_trim))
        fid.write("trimLeft,{}\\n".format(left_trim))
        fid.write("truncLen,{}\\n".format(length_trim))


if __name__ == "__main__":
    FORWARD = "${qual_summary}"
    quality_analysis(FORWARD)



"""
    Module that defines the environment setup for processes
"""

import pathlib
from typing import Iterable, Optional

from ..pipelines import Command
from ..logging import LOG


ENV_DIR = pathlib.Path(__file__).parent.parent / "pipelines/envs"


class Environments:
    """
    A class that creates, lists and loads conda environments

    Parameters
    ----------

    Attributes
    ----------
    configs : List[pathlib.Path]
        The list of locations of the environment config files
    env_names : List[str]
        The list of names of environments
    """

    def __init__(self) -> None:
        self.configs = list(ENV_DIR.glob("**/env.yml"))
        self.env_names = [f"{c.parent.stem}" for c in self.configs]

    def init(self, env: Optional[str] = None) -> Iterable[Command]:
        """
        Initialize the requested conda environment

        Parameters
        ----------
        env : Optional[str]
            The name of the conda environment to initialize
            If None then all the listed conda environments will be initialized
            Default value is None

        Yields
        ------
        Command
            The currently running initialization command
        """
        if env is None:
            for config, env_name in zip(self.configs, self.env_names):
                LOG.logger.info(f"Initializing {env_name} environment")
                cmd_str = f"conda env create -f {config} -n {env_name}"
                init_cmd = Command(cmd_str, profile="local", timeout=10000)
                init_cmd.run()
                yield init_cmd
        elif env in self.env_names:
            ind = self.env_names.index(env)
            config = self.configs[ind]
            env_name = self.env_names[ind]
            LOG.logger.info(f"Initializing {env_name} environment")
            cmd_str = f"conda env create -f {config} -n {env_name}"
            init_cmd = Command(cmd_str, profile="local", timeout=10000)
            init_cmd.run()
            yield init_cmd
        elif env not in self.env_names:
            raise ValueError(f"{env} not a supported environment")

    def post_install(self, env: Optional[str] = None) -> Iterable[Command]:
        """
        Run any post installation scripts for environment setup

        Parameters
        ----------
        env : Optional[str]
            The name of the conda environment to setup
            If None then all the listed conda environments will be initialized
            Default value is None

        Yields
        ------
        Command
            The currently running post_install command
        """
        if env is None:
            post_scripts = list(ENV_DIR.glob("**/post_install.sh"))
        else:
            post_scripts = list(ENV_DIR.glob(f"**/{env}/post_install.sh"))
        for script in post_scripts:
            cmd_str = f"bash {script}"
            LOG.logger.info(f"Running post_install for {env} environment")
            post_cmd = Command(cmd_str, profile="local", timeout=10000)
            post_cmd.run()
            yield post_cmd

    def load(self, env: str) -> None:
        """
        Load the requested conda environment

        Parameters
        ----------
        env : str
            The name of the conda environment to load

        """
        if env not in self.env_names:
            raise ValueError(f"{env} not a supported environment")
        ind = self.env_names.index(env)
        env_name = self.env_names[ind]
        cmd_str = f"source activate {env_name}"
        LOG.logger.info(f"Loading {env} environment")
        load_cmd = Command(cmd_str, profile="local", timeout=10000)
        load_cmd.run()



"""
    Module that defines the `Otu` objects and methods to manipulate it
"""

import json
import pathlib
from typing import Callable, Dict, Hashable, Iterable, List, Optional, Tuple

from biom import Table
from biom.util import biom_open
import numpy as np
import pandas as pd

from ..validation import OtuValidator, BiomType, SamplemetaType, ObsmetaType
from .lineage import Lineage


Filterfun = Callable[[np.ndarray, str, dict], bool]
Hashfun = Callable[[str, dict], Hashable]


class Otu:
    """
    An object that represents the OTU counts table

    Parameters
    ----------
    otu_data : Table
        `biom.Table` object containing OTU data
    sample_metadata : pd.DataFrame, optional
        `pd.DataFrame` containing metadata for the samples
    obs_metadata : pd.DataFrame,  optional
        `pd.DataFrame` containing metadata for the observations (OTUs)

    Attributes
    ----------
    otu_data : biom.Table
        OTU counts table in the `biom.Table` format
    sample_metadata : pd.DataFrame
        Metadata for the samples
    obs_metadata :  pd.DataFrame
        Lineage data for the observations (OTUs)
    tax_level : str
        The taxonomy level of the current Otu instance

    Notes
    -----
    All methods that manipulate the Otu object return new objects
    """

    def __init__(
        self,
        otu_data: Table,
        sample_metadata: Optional[pd.DataFrame] = None,
        obs_metadata: Optional[pd.DataFrame] = None,
    ) -> None:
        if not isinstance(otu_data, Table):
            raise TypeError("Otu data must be of type `biom.Table`")
        otu_data_copy = otu_data.copy()
        if sample_metadata is not None:
            samplemeta_type = SamplemetaType()
            samplemeta_type.validate(sample_metadata)
            otu_data_copy.add_metadata(
                sample_metadata.to_dict(orient="index"), axis="sample"
            )
        if obs_metadata is not None:
            obsmeta_type = ObsmetaType()
            obsmeta_type.validate(obs_metadata)
            otu_data_copy.add_metadata(
                obs_metadata.to_dict(orient="index"), axis="observation"
            )
        biom_type = BiomType()
        biom_type.validate(otu_data_copy)
        self.otu_data = otu_data_copy

    def __repr__(self) -> str:
        n_obs, n_samples = self.otu_data.shape
        return f"<Otu {n_obs}obs x {n_samples}samples>"

    @classmethod
    def load_data(
        cls,
        otu_file: str,
        meta_file: Optional[str] = None,
        tax_file: Optional[str] = None,
        dtype: str = "biom",
        ext: Optional[str] = None,
    ) -> "Otu":
        """
        Load data from files into the `Otu` class instance

        Parameters
        ----------
        otu_file : str
            The path to the `OTU` counts file
        meta_file : str, optional
            The path to the sample metadata file
        tax_file : str, optional
            The path to the taxonomy file
        dtype : {'biom', 'tsv'}
            The type of OTU file that is input
        ext : str, optional
            The extension of the file if other than supported extensions
            Supported extensions:
            - 'tsv' dtype: 'tsv', 'txt', 'counts'
            - 'biom' dtype: 'biom', 'hdf5'

        Returns
        -------
        Otu
            An instance of the `Otu` class
        """
        otu_validator = OtuValidator(dtype=dtype, ext=ext)
        otu_path = pathlib.Path(otu_file)
        if otu_path.exists():
            meta_path = pathlib.Path(meta_file) if meta_file else meta_file
            tax_path = pathlib.Path(tax_file) if tax_file else tax_file
            otu_data = otu_validator.load_validate(otu_path, meta_path, tax_path)
        else:
            raise FileNotFoundError("Missing input files")
        return cls(otu_data)

    @property
    def sample_metadata(self) -> pd.DataFrame:
        """
        Metadata for the samples

        Returns
        -------
        pd.DataFrame
        """
        return self.otu_data.metadata_to_dataframe("sample")

    @property
    def obs_metadata(self) -> pd.DataFrame:
        """Lineage data for the observations (OTUs)"""
        obs_metadata = self.otu_data.metadata_to_dataframe("observation")
        lineage = list(Lineage._fields)
        n_tax_levels = len(set(obs_metadata.columns) & set(lineage))
        lineage_columns = lineage[:n_tax_levels]
        extra_columns = list(set(obs_metadata.columns) - set(lineage_columns))
        if extra_columns:
            columns: List[str] = lineage_columns + extra_columns
        else:
            columns = lineage_columns
        return obs_metadata[columns]

    @property
    def tax_level(self) -> str:
        """
        Returns the taxonomy level of the Otu instance

        Returns
        -------
        str
            The lowest taxonomy defined in the Otu instance
        """
        obs_metadata = self.otu_data.metadata_to_dataframe("observation")
        lineage = list(Lineage._fields)
        n_tax_levels = len(set(obs_metadata.columns) & set(lineage))
        return Lineage._fields[n_tax_levels - 1]

    def filter(
        self,
        ids: Optional[Iterable[str]] = None,
        func: Optional[Filterfun] = None,
        axis: str = "observation",
    ) -> "Otu":
        """
        Filter Otu instance based on ids or func

        Parameters
        ----------
        ids : Iterable[str], optional
            An iterable of ids to keep.
            If ids are not supplied then func must be supplied
        func : Callable[[np.ndarray, str, dict], bool], optional
            A function that takes in (values, id_, md) and returns a bool
            If func is not supplied then ids must be supplied
            If both ids and func are supplied then ids are used
        axis : {'sample', 'observation'}, optional
            The axis along which to filter the Otu instance
            Default value is 'observation'

        Returns
        -------
        Otu
            Filtered Otu instance
        """
        if ids is not None:
            otu_filtered = self.otu_data.filter(ids, inplace=False, axis=axis)
        elif func:
            otu_filtered = self.otu_data.filter(func, inplace=False, axis=axis)
        else:
            raise TypeError("Either ids or func must be supplied")
        return Otu(otu_filtered)

    def normalize(self, axis: str = "sample", method: str = "norm") -> "Otu":
        """
        Normalize the OTU table along the provided axis

        Parameters
        ----------
        axis : {'sample', 'observation'}, optional
            Axis along which to normalize the OTU table
            Default is 'sample'
        method: {'norm', 'rarefy', 'css'}
            Normalization method to use

        Returns
        -------
        Otu
            Otu instance which is normalized along the given axis
        """
        if method == "norm":
            norm_otu = self.otu_data.norm(axis=axis, inplace=False)
        elif method == "rarefy":
            raise NotImplementedError("This method is not implemented yet")
        elif method == "css":
            raise NotImplementedError("This method is not implemented yet")
        else:
            raise ValueError(
                "Invalid method. Supported methods are {'norm', 'rarefy', 'css'}"
            )
        return Otu(norm_otu)

    def is_norm(self, axis: str = "sample") -> bool:
        """
        Returns true if the Otu instance has been normalized
        """
        otu_data = self.otu_data.to_dataframe()
        if axis == "sample":
            return bool(np.isclose(otu_data.sum(axis=0), 1.0).all())
        if axis == "observation":
            return bool(np.isclose(otu_data.sum(axis=1), 1.0).all())
        raise ValueError("Axis must of either {'sample' or 'observation'}")

    def rm_sparse_samples(self, count_thres: int = 500) -> "Otu":
        """
        Remove samples with read counts less than `count_thres`

        Parameters
        ----------
        count_thres : int, optional
            Counts threshold below which samples are rejected
            Default value is 500

        Returns
        -------
        Otu
            Otu instance with low count samples removed

        Raises
        ------
        ValueError
            If Otu instance is normalized
        """
        if self.is_norm():
            raise ValueError(
                "Otu instance is normalized and hence will not work with this method"
            )
        filt_fun = lambda val, *_: round(val.sum()) >= count_thres
        new_otu = self.otu_data.filter(filt_fun, axis="sample", inplace=False)
        return Otu(new_otu)

    def rm_sparse_obs(
        self,
        prevalence_thres: float = 0.05,
        abundance_thres: float = 0.01,
        obssum_thres: int = 100,
    ) -> "Otu":
        """
        Remove observations with prevalence < `prevalence_thres` and abundance < `abundance_thres`

        Parameters
        ----------
        prevalence_thres : float
            Minimum fraction of samples the observation must be present in in order to be accepted
        abundance_thres : float
            Minimum observation count fraction in a sample needed in order to be accepted
        obssum_thres : int
            The theshold applied to the sum of observations for each row

        Returns
        -------
        Otu
            Otu instance with bad observations removed
        """
        prevalence_obssum_filter = lambda val, *_: (
            (val.astype(int).astype(bool).mean()) >= prevalence_thres
        ) and ((val.astype(int).sum()) >= obssum_thres)
        otu_prevalence_obssum_filtered = self.otu_data.filter(
            prevalence_obssum_filter, axis="observation", inplace=False
        )
        otu_df = otu_prevalence_obssum_filtered.to_dataframe()
        if otu_df.apply(pd.api.types.is_sparse).any():
            otu_rel_abund = (otu_df / otu_df.sum(axis=0)).sparse.to_dense()
        else:
            otu_rel_abund = otu_df / otu_df.sum(axis=0)
        ind_above_thres = otu_rel_abund.index[
            (otu_rel_abund > abundance_thres).any(axis=1)
        ]
        new_otu = self.otu_data.filter(
            ind_above_thres, axis="observation", inplace=False
        )
        ind_below_thres = set(self.otu_data.ids("observation")) - set(ind_above_thres)
        otu_sparse_obs = self.otu_data.filter(
            ind_below_thres, axis="observation", inplace=False
        )
        new_row = Table(
            otu_sparse_obs.sum(axis="sample"),
            ["otu_merged"],
            self.otu_data.ids(axis="sample"),
        )
        tax_level = self.tax_level
        random_row_metadata = dict(self.otu_data.metadata(axis="observation")[0])
        new_row.add_metadata(
            {
                "otu_merged": {
                    **random_row_metadata,
                    **Lineage("Unclassified").to_dict(tax_level),
                }
            },
            axis="observation",
        )
        final_otu = new_otu.concat([new_row], axis="observation")
        return Otu(final_otu)

    def partition(self, axis: str, func: Hashfun) -> Iterable[Tuple[str, "Otu"]]:
        """
        Partition the Otu instance based on the func and axis

        Parameters
        ----------
        axis : str
            The axis on which to partition
        func : Callable[[str, dict], Hashable]
            The function that takes in (id, metadata) and returns a hashable

        Returns
        -------
        Iterable[Tuple[str, Otu]]
            An iterable of tuples - ('label', Otu)

        Notes
        -----
        1. To group by lineage "level" use:
            func = lambda id_, md: Lineage(**md).get_superset(level)
        """
        if axis == "observation" and self.is_norm(axis="sample"):
            raise ValueError(
                "Cannot partition sample normalized Otu instance on observation"
            )
        if axis == "sample" and self.is_norm(axis="observation"):
            raise ValueError(
                "Cannot partition observation normalized Otu instance on sample"
            )
        partitions = self.otu_data.partition(func, axis=axis)
        for label, table in partitions:
            yield label, Otu(table)

    def collapse_taxa(self, level: str) -> Tuple["Otu", Dict[str, List[str]]]:
        """
        Collapse Otu instance based on taxa

        Parameters
        ----------
        level : str
            The tax level of the collapsed table
            This will also be used as the prefix for the unique ids

        Returns
        -------
        Tuple[Otu, dict]
            Collapsed Otu instance
        """
        obs_metadata_cols = list(self.obs_metadata)
        lineage_cols = list(Lineage._fields)
        unwanted_obs_metadata_cols = list(set(obs_metadata_cols) - set(lineage_cols))
        otu_data_copy = self.otu_data.copy()
        otu_data_copy.del_metadata(keys=unwanted_obs_metadata_cols, axis="observation")
        obs_metadata_copy = self.obs_metadata.drop(unwanted_obs_metadata_cols, axis=1)
        if level not in Lineage._fields:
            raise ValueError(f"level must be one of {Lineage._fields}")
        cfunc = lambda id_, md: str(Lineage(**md).get_superset(level))
        otu_collapse = otu_data_copy.collapse(
            cfunc, axis="observation", norm=False, include_collapsed_metadata=True
        )
        curr_ids = otu_collapse.ids(axis="observation")
        children_group_list = [
            i["collapsed_ids"] for i in otu_collapse.metadata(axis="observation")
        ]
        children_groups = dict(zip(curr_ids, children_group_list))
        otu_collapse.del_metadata(axis="observation")
        afunc = lambda x: pd.Series(
            Lineage(**x.to_dict()).get_superset(level).to_dict(level)
        )
        obs_collapse = obs_metadata_copy.apply(afunc, axis=1)
        gfunc = lambda x: str(Lineage(**x.to_dict()))
        obs_collapse.index = obs_collapse.apply(gfunc, axis=1)
        obs_collapse.drop_duplicates(inplace=True)
        otu_collapse.add_metadata(
            obs_collapse.to_dict(orient="index"), axis="observation"
        )
        obs_dict = json.loads(
            otu_collapse.metadata_to_dataframe("observation").to_json(orient="split")
        )
        unq_id_map = {k: f"{level}_{i}" for i, k in enumerate(curr_ids)}
        obs_ids = [unq_id_map[i] for i in obs_dict["index"]]  # get ordered unique ids
        children_dict = {unq_id_map[k]: v for k, v in children_groups.items()}
        sample_ids = otu_collapse.ids(axis="sample")
        observation_metadata = otu_collapse.metadata_to_dataframe(
            axis="observation"
        ).to_dict(orient="records")
        sample_metadata = otu_collapse.metadata_to_dataframe(axis="sample").to_dict(
            orient="records"
        )
        new_table = Table(
            otu_collapse.matrix_data,
            obs_ids,
            sample_ids,
            observation_metadata=observation_metadata,
            sample_metadata=sample_metadata,
        )
        return Otu(new_table), children_dict

    def write(
        self, base_name: str, fol_path: str = "", file_type: str = "biom"
    ) -> None:
        """
        Write Otu instance object to required file_type

        Parameters
        ----------
        base_name : str
            The base name without extension to be used for the files
        fol_path : str, optional
            The folder where the files are to be written
            Default is current directory
        file_type : {'tsv', 'biom'}, optional
            The type of file data is to be written to
            Default is 'biom'
        """
        folder = pathlib.Path(fol_path)
        if not folder.exists():
            folder.mkdir()
        if file_type == "biom":
            fname = base_name + ".biom"
            fpath = str(folder / fname)
            with biom_open(fpath, "w") as fid:
                self.otu_data.to_hdf5(fid, "Constructed using micone")
        elif file_type == "tsv":
            otu_name = base_name + "_otu.tsv"
            otu_path = folder / otu_name
            with open(otu_path, "w") as fid:
                # NOTE: delete header comment from output
                data = self.otu_data.to_tsv().split("\n", 1)[-1]
                fid.write(data)
            sample_metadata_name = base_name + "_sample_metadata.tsv"
            sample_metadata_path = folder / sample_metadata_name
            self.sample_metadata.to_csv(sample_metadata_path, sep="\t", index=True)
            obs_metadata_name = base_name + "_obs_metadata.csv"
            obs_metadata_path = folder / obs_metadata_name
            self.obs_metadata.to_csv(obs_metadata_path, index=True)
        else:
            raise ValueError("Supported file types are 'tsv' and 'biom'")



#!/usr/bin/env python3

from qiime2 import Artifact
from qiime2.plugins import demux


def main(sequence_artifact, seq_samplesize, output_dir):
    demux_artifact = Artifact.load(sequence_artifact)
    demux_viz = demux.visualizers.summarize(demux_artifact, n=seq_samplesize)
    demux_viz.visualization.export_data(output_dir)


if __name__ == "__main__":
    SEQUENCE_ARTIFACT = "${sequence_artifact}"
    SEQ_SAMPLESIZE = ${seq_samplesize}
    OUTPUT_DIR = "output"
    main(SEQUENCE_ARTIFACT, SEQ_SAMPLESIZE, OUTPUT_DIR)



"""
    Custom spinner class
"""

from typing import Optional

from halo import Halo
from click import echo


class Spinner:
    """
    A custom spinner class that makes use of `halo` and `click`

    Parameters
    ----------
    text : str
    spinner : str

    Attributes
    ----------
    interactive : bool
        True if the `Spinner` instance is interactive
    text : str
        The text currently in the spinner
    """

    def __init__(self, text: str, spinner: str, interactive: bool = True) -> None:
        self.interactive = interactive
        self._text = text
        if self.interactive:
            self._spinner = Halo(text=self._text, spinner=spinner)
        else:
            self._spinner = None

    @property
    def text(self) -> str:
        """ The text in the spinner """
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        self._text = text
        if self.interactive:
            self._spinner.text = self._text
        else:
            echo(self._text)

    def start(self) -> None:
        """ Start the spinner """
        if self.interactive:
            self._spinner.start()

    def succeed(self, text: Optional[str] = None) -> None:
        """ Display text on success """
        if text:
            self._text = text
        if self.interactive:
            self._spinner.succeed(self.text)
        else:
            echo(self.text)

    def fail(self, text: Optional[str] = None) -> None:
        """ Display text on failure """
        if text:
            self._text = text
        if self.interactive:
            self._spinner.fail(self.text)
        else:
            echo(self.text)

    def stop(self) -> None:
        """ Stop the spinner """
        if self.interactive:
            self._spinner.stop()
        self._text = ""



"""
    Console script for micone
"""

import pathlib
import pprint
from typing import List

import click

from .logging import LOG
from .setup import Environments
from .utils import Spinner
from .validation import check_results


@click.group()
@click.option(
    "--log", "-l", default=True, type=bool, help="Flag to turn on/off logging"
)
@click.option(
    "--interactive", "-i", default=True, type=bool, help="Flag to turn on/off spinner"
)
@click.pass_context
def cli(ctx, log: bool, interactive: bool):
    """Main entry point to micone"""
    spinner = Spinner(text="Starting up...", spinner="dots", interactive=interactive)
    spinner.start()
    ctx.obj["SPINNER"] = spinner
    spinner.succeed("Successfully initialized micone")
    if log:
        LOG.enable()
        click.secho(f"Log file is at {LOG.path}")


@cli.command()
@click.option(
    "--env",
    "-e",
    default=None,
    type=str,
    help="The environment to initialize. By default will initialize all",
)
@click.pass_context
def init(ctx, env: str):
    """Initialize the package and environments"""
    spinner = ctx.obj["SPINNER"]
    environments = Environments()
    for env_cmd in environments.init(env):
        spinner.start()
        spinner.text = f"Initializing environment: {env_cmd}"
        env_cmd.wait()
        env_cmd.log()
        if env_cmd.status == "failure":
            spinner.fail(f"{env_cmd} Failed")
        elif env_cmd.status == "success":
            spinner.succeed(f"{env_cmd} Passed")
    for post_cmd in environments.post_install(env):
        spinner.start()
        spinner.text = f"Running post installation: {post_cmd}"
        post_cmd.wait()
        post_cmd.log()
        if post_cmd.status == "failure":
            spinner.fail(f"{post_cmd} Failed")
        elif post_cmd.status == "success":
            spinner.succeed(f"{post_cmd} Passed")
    click.secho(f"Log file is at {LOG.path}")
    LOG.cleanup()


@cli.command()
@click.option(
    "--profile",
    "-p",
    default="local",
    type=str,
    help="The execution profile. Either 'local' or 'sge'",
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="The config file that defines the pipeline run",
)
@click.option(
    "--output_location",
    "-o",
    type=click.Path(exists=True),
    default=None,
    help="The base output location to store pipeline results",
)
@click.option(
    "--base_dir",
    "-b",
    type=click.Path(exists=True),
    default=None,
    help="The location of base directory for input files",
)
@click.option(
    "--max_procs",
    "-m",
    type=click.INT,
    default=4,
    help="Maximum number of processes allowed to run in parallel",
)
@click.option(
    "--resume",
    is_flag=True,
    help="The flag to determine whether a previous execution is resumed",
)
@click.pass_context
def run(
    ctx,
    profile: str,
    config: click.Path,
    output_location: click.Path,
    base_dir: click.Path,
    max_procs: int,
    resume: bool,
):
    """Run the pipeline"""
    spinner = ctx.obj["SPINNER"]
    pipeline = Pipeline(
        str(config), profile, str(base_dir), resume, output_location=output_location
    )
    spinner.start()
    spinner.text = "Starting pipeline execution"
    try:
        for process in pipeline.run(max_procs=max_procs):
            if pipeline.process_queue is None:
                raise ValueError("Pipeline process queue is empty")
            process_list = " and ".join(
                [proc.id_.split(".", 2)[-1] for proc in pipeline.process_queue]
            )
            if resume and process.io_exist:
                spinner.start()
                spinner.text = f"Executing {process_list}"
                spinner.succeed(f"Resumed {process}")
            spinner.start()
            spinner.text = f"Executing {process_list}"
            updated_processes = pipeline.wait()
            for proc in updated_processes:
                proc.log()
                if proc.status == "success":
                    spinner.succeed(f"Completed {proc}")
                else:
                    spinner.fail(f"Failed to execute {proc}")
    finally:
        LOG.cleanup()


@cli.command()
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="The config file that defines the pipeline run",
)
@click.option(
    "--files", "-f", multiple=True, help="Files can be logs, configs, work, results"
)
@click.pass_context
def clean(ctx, config: click.Path, files: List[str]):
    """Clean files from a pipeline run in the current directory.
    This command irrevesibly deletes content, use with caution
    """
    click.confirm(
        f"Are you sure you want to delete {files} in current directory?", abort=True
    )
    spinner = ctx.obj["SPINNER"]
    spinner.start()
    spinner.text = "Cleaning up pipeline files"
    pipeline = Pipeline(str(config), "local")
    pipeline.clean(files)
    spinner.succeed("Completed cleanup")


@cli.command()
@click.option(
    "--dir", "-d", type=click.Path(exists=True), help="The pipeline execution directory"
)
@click.option(
    "--procs",
    "-p",
    type=click.INT,
    default=4,
    help="Number of processors to use",
)
@click.pass_context
def validate_results(ctx, dir: click.Path, procs: int):
    """Check results of the pipeline execution"""
    pipeline_dir = pathlib.Path(str(dir))
    spinner = ctx.obj["SPINNER"]
    spinner.start()
    spinner.text = "Checking pipeline execution"
    (
        pipeline_results,
        trace_results,
        biom_results,
        network_results,
        output_results,
    ) = check_results(pipeline_dir, procs)
    if pipeline_results == "success":
        spinner.succeed("Pipeline execution was successful")
    else:
        spinner.fail("Pipeline execution was unsuccessful")
    spinner.start()
    if not biom_results.get("fail", None):
        spinner.succeed("Biom files had no errors")
    else:
        spinner.fail("Biom files had errors")
        pprint.pprint(biom_results)
    if not network_results.get("fail", None):
        spinner.succeed("Network files had no errors")
    else:
        spinner.fail("Network files had errors")
        pprint.pprint(network_results)
    if not output_results["fail"]:
        spinner.succeed("All modules were executed completely")
    else:
        spinner.fail("All modules were not executed completely")
        pprint.pprint(output_results)


def main():
    cli(obj={})


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import pathlib

from biom import Table
from biom.util import biom_open
import pandas as pd


def make_repseqs(seqs, output_file):
    with open(output_file, "w") as fid:
        for i, seq in enumerate(seqs):
            fid.write(f">Seq{i}\\n")
            fid.write(seq)
            fid.write("\\n")


def make_biom(seq_table, output_file):
    new_table = Table(seq_table.values, list(seq_table.index), list(seq_table.columns))
    with biom_open(output_file, "w") as fid:
        new_table.to_hdf5(fid, "Constructucted by micone in dada2 pipeline")


def main(seqtable_file, meta_id, sample_metadata):
    seq_table = pd.read_table(seqtable_file, index_col=0)
    make_repseqs(list(seq_table.index), f"{meta_id}_unhashed_rep_seqs.fasta")
    make_biom(seq_table, f"{meta_id}_unhashed_otu_table.biom")
    sample_metadata_path = pathlib.Path(sample_metadata)
    sample_metadata_path.rename(f"{meta_id}_sample_metadata.tsv")


if __name__ == "__main__":
    SEQ_TABLE_FILE = "${seq_table_file}"
    SAMPLE_METADATA = "${samplemetadata_files}"
    META_ID = "${meta.id}-${meta.run}"
    main(SEQ_TABLE_FILE, META_ID, SAMPLE_METADATA)



#!/usr/bin/env python3

import pathlib
import multiprocessing as mp

import numpy as np
import pandas as pd
from scipy.stats import spearmanr


def spearman(otu_file, output_file):
    otu_table = pd.read_table(otu_file, index_col=0)
    data = otu_table.values
    n = otu_table.shape[0]
    index = otu_table.index
    corr_data = np.zeros((n, n), dtype=np.float)
    for i in range(n):
        for j in range(i, n):
            if i == j:
                corr_data[i][j] = 1.0
            else:
                corr, _ = spearmanr(data[i, :], data[j, :])
                corr_data[i][j] = corr
                corr_data[j][i] = corr
    corr_table = pd.DataFrame(data=corr_data, index=index, columns=index)
    corr_table.fillna(value=0.0, inplace=True)
    corr_table.to_csv(output_file, sep="\\t", index=True, float_format="%.4f")


def main(id_, otu_file, bootstrap_files, ncpus):
    output_file = f"{id_}_corr.tsv"
    args = [(otu_file, output_file)]
    for i, bootstrap_file in enumerate(bootstrap_files):
        output_file = bootstrap_file.name.replace("_otu.boot", "_corr.boot")
        args.append((bootstrap_file, output_file))
    with mp.Pool(processes=ncpus) as pool:
        pool.starmap(spearman, args)


if __name__ == "__main__":
    ID_ = "${meta.id}"
    OTU_FILE = pathlib.Path("${otu_file}")
    BOOTSTRAP_FILES = pathlib.Path().glob("*_otu.boot")
    NCPUS = int("${ncpus}")
    main(ID_, OTU_FILE, BOOTSTRAP_FILES, NCPUS)



#!/usr/bin/env python3

import pathlib
from typing import List

from micone import Network, NetworkGroup


def main(
    base_name: str,
    network_files: List[pathlib.Path],
    method: str,
    parameter: float,
    pvalue_filter: bool,
    interaction_filter: bool,
    id_field: str,
) -> None:
    networks: List[Network] = []
    for network_file in network_files:
        networks.append(Network.load_json(str(network_file)))
    network_group = NetworkGroup(networks, id_field=id_field)
    pathlib.Path("consensus/").mkdir(parents=True, exist_ok=True)
    cids = list(range(len(network_group.contexts)))
    filtered_network_group = network_group.filter(
        pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
    )
    consensus_network_group = filtered_network_group.get_consensus_network(
        cids, method=method, parameter=parameter
    )
    consensus_network_group.write("consensus/" + base_name + "_network.json")


if __name__ == "__main__":
    BASE_NAME = "consensus"
    METHOD = "${method}"
    PARAMETER = float("${parameter}")
    PVALUE_FILTER = True if "${pvalue_filter}" == "true" else False
    INTERACTION_FILTER = True if "${interaction_filter}" == "true" else False
    ID_FIELD = "${id_field}"
    NETWORK_FILES = list(pathlib.Path().glob("*_network.json"))
    main(
        BASE_NAME,
        NETWORK_FILES,
        METHOD,
        PARAMETER,
        PVALUE_FILTER,
        INTERACTION_FILTER,
        ID_FIELD,
    )



#!/usr/bin/env python3

import pathlib
import os
import subprocess


def convert(data):
    sample_name, file_name = data
    fasta_file = "fasta/" + sample_name + ".fasta"
    with open(fasta_file, "w") as fid:
        cmd = ["seqtk", "seq", "-a", file_name]
        subprocess.call(cmd, stdout=fid)
    return fasta_file


def merge(fasta_files, output_file):
    with open(output_file, "w") as wid:
        for fasta_file in fasta_files:
            with open(fasta_file) as rid:
                sample_name = pathlib.Path(fasta_file).stem
                count = 0
                for line in rid:
                    if line.startswith(">"):
                        wid.write(
                            ">{0}_{1} {2}".format(sample_name, str(count), line[1:])
                        )
                    else:
                        wid.write(line)
                    count += 1


def main(manifest_file, dataset_name):
    os.mkdir("fasta")
    data = []
    with open(manifest_file) as fid:
        _ = next(fid).strip().split(",")
        for line in fid:
            if not line.startswith("#"):
                sample_name, file_location, _ = line.strip().split(",")
                data.append((sample_name, file_location.rsplit("/")[-1]))
    fasta_files = map(convert, data)
    merged_fasta_file = dataset_name + ".fasta"
    merge(fasta_files, merged_fasta_file)


if __name__ == "__main__":
    DATASET_NAME = "${meta.id}"
    MANIFEST_FILE = "${manifest_file}"
    main(MANIFEST_FILE, DATASET_NAME)



from .otu_schema import BiomType, SamplemetaType, ObsmetaType
from .network_schema import (
    InteractionmatrixType,
    CorrelationmatrixType,
    PvaluematrixType,
    MetadataModel,
    ChildrenmapType,
    NodesModel,
    LinksModel,
    NetworkmetadataModel,
    ElistType,
)
from .otu_validator import OtuValidator
from .execution_validator import check_results



#!/usr/bin/env python3

import hashlib
import pathlib

from biom import load_table, Table
from biom.util import biom_open
from Bio import SeqIO
from Bio.SeqIO import FastaIO
import pandas as pd


def hash_function(seq):
    return hashlib.md5(seq.encode("utf-8")).hexdigest()


def hash_otu_table(unhashed_otu_table, seqid_hash_dict):
    table = load_table(unhashed_otu_table)
    df = table.to_dataframe(dense=True)
    seq_ids = [seqid_hash_dict[i] for i in df.index]
    df.index = seq_ids
    return df


def hash_rep_seqs(unhashed_rep_seqs, rep_seq_ids):
    seqs_all = list(SeqIO.parse(unhashed_rep_seqs, "fasta"))
    seqs = []
    seqid_hash_dict = dict()
    for seq in seqs_all:
        hash_ = hash_function(str(seq.seq))
        seqid_hash_dict[seq.id] = hash_
        seq.id = hash_function(str(seq.seq))
        seq.description = ""
        seq.name = ""
        if seq.id not in rep_seq_ids:
            rep_seq_ids.add(hash_)
            seqs.append(seq)
    return seqs, seqid_hash_dict


def hashing(unhashed_otu_table_list, unhashed_rep_seqs_list, sample_metadata_list):
    otu_df_list = []
    rep_seq_ids = set()
    seqs = []
    for unhashed_rep_seqs, unhashed_otu_table in zip(
        unhashed_rep_seqs_list, unhashed_otu_table_list
    ):
        seq, seqid_hash_dict = hash_rep_seqs(unhashed_rep_seqs, rep_seq_ids)
        seqs.extend(seq)
        otu_df_list.append(hash_otu_table(unhashed_otu_table, seqid_hash_dict))
    # Merge OTU table
    otu_df = pd.concat(otu_df_list, join="outer", axis=1)
    otu_df.fillna(0.0, inplace=True)
    otu_table = Table(otu_df.values, list(otu_df.index), list(otu_df.columns))
    otu_table_ids = set(otu_df.index)
    assert otu_table_ids == rep_seq_ids
    assert len(otu_df.index) == len(rep_seq_ids)
    # Merge rep seqs
    # Merge sample metadata
    sample_metadata = pd.concat(
        [pd.read_csv(s, sep="\\t") for s in sample_metadata_list]
    )

    # Write files
    sample_metadata.to_csv("sample_metadata.tsv", sep="\\t", index=False)
    with biom_open("otu_table.biom", "w") as fid:
        otu_table.to_hdf5(fid, "Constructed using qiime2 clustering")
    with open("rep_seqs.fasta", "w") as fid:
        fasta_writer = FastaIO.FastaWriter(fid, wrap=None)
        fasta_writer.write_file(seqs)


if __name__ == "__main__":
    UNHASHED_OTU_TABLE_LIST = sorted(pathlib.Path().glob("*unhashed_otu_table.biom"))
    UNHASHED_REP_SEQS_LIST = sorted(pathlib.Path().glob("*unhashed_rep_seqs.fasta"))
    SAMPLE_METADATA_LIST = sorted(pathlib.Path().glob("*sample_metadata.tsv"))
    hashing(UNHASHED_OTU_TABLE_LIST, UNHASHED_REP_SEQS_LIST, SAMPLE_METADATA_LIST)



"""
    Module containing methods that convert taxonomy metadata into various formats
"""

import csv
import pathlib
from typing import Dict, List

import pandas as pd

from ..validation.otu_schema import ObsmetaType
from ..main import Lineage


def qiime2_to_default(in_file: pathlib.Path, out_file: pathlib.Path) -> None:
    """
    Convert Taxonomy metadata file from qiime2 to default format

    Parameters
    ----------
    in_file : pathlib.Path
        The path to the qiime2 formatted input file
    out_file : pathlib.Path
        The path to the default format output file
    """
    with open(in_file, "r") as fid:
        csv_reader = csv.reader(fid, delimiter="\t")
        data_dict: Dict[str, List[str]] = {key: [] for key in ObsmetaType._req_keys}
        data_dict["ID"] = []
        header, *rest = csv_reader
        ekey = ObsmetaType._extra_keys[0]  # the Confidence value
        if ekey in header:
            data_dict[ekey] = []
        for line in rest:
            if line[0].startswith("#q2:types"):
                continue
            lineage = Lineage.from_str(line[header.index("Taxon")])
            data_dict["ID"].append(line[0])
            for rkey in ObsmetaType._req_keys:
                data_dict[rkey].append(getattr(lineage, rkey))
            if ekey in data_dict:
                data_dict[ekey].append(line[header.index(ekey)])
    df = pd.DataFrame.from_dict(data_dict)
    df.set_index("ID", inplace=True, verify_integrity=True)
    df.to_csv(out_file, index=True)


def qiime1_to_default(in_file: pathlib.Path, out_file: pathlib.Path) -> None:
    """
    Convert Taxonomy metadata file from qiime1 to default format

    Parameters
    ----------
    in_file : pathlib.Path
        The path to the qiime1 formatted input file
    out_file : pathlib.Path
        The path to the default format output file
    """
    pass


def default_to_qiime2(in_file: pathlib.Path, out_file: pathlib.Path) -> None:
    """
    Convert Taxonomy metadata file from default format to qiime2

    Parameters
    ----------
    in_file : pathlib.Path
        The path to the default format input file
    out_file : pathlib.Path
        The path to the qiime2 formatted output file
    """
    pass


CONVERTERS = {
    ("qiime2", "default"): qiime2_to_default,
    ("qiime1", "default"): qiime1_to_default,
    ("default", "qiime2"): default_to_qiime2,
}



"""
    Module that implements the `Lineage` class and  methods to work with taxonomy data
"""


from collections import namedtuple
from typing import Dict, Tuple
from warnings import warn

from ete3 import NCBITaxa

from ..logging import LOG

BaseLineage = namedtuple("Lineage", "Kingdom Phylum Class Order Family Genus Species")


class Lineage(BaseLineage):
    """
    `NamedTuple` that stores the lineage of a taxon and methods to interact with it

    Attributes
    ----------
    Kingdom: str
    Phylum: str
    Class: str
    Order: str
    Family: str
    Genus: str
    Species: str
    """

    def __new__(
        cls,
        Kingdom: str = "",
        Phylum: str = "",
        Class: str = "",
        Order: str = "",
        Family: str = "",
        Genus: str = "",
        Species: str = "",
    ) -> "Lineage":
        tax_order = [Kingdom, Phylum, Class, Order, Family, Genus, Species]
        empty = [i for i, tax in enumerate(tax_order) if tax == ""]
        if empty and (len(tax_order) - empty[0] != len(empty)):
            warn(
                RuntimeWarning(
                    f"Lower levels should not be filled if higher levels are empty: {tax_order}"
                )
            )
        norm_taxa = [cls._normalize_tax(i) for i in tax_order]
        cls._ncbi = NCBITaxa()
        return super().__new__(cls, *norm_taxa)

    @staticmethod
    def _normalize_tax(tax: str) -> str:
        """
        Normalize taxonomy name by removing unwanted characters

        Parameters
        ----------
        tax : str

        Returns
        -------
        str
            Normalized taxonomy name
        """
        return (
            tax.strip()
            .replace("[", "")
            .replace("]", "")
            .replace("'", "")
            .replace("=", "")
        )

    def __sub__(self, other: "Lineage") -> "Lineage":
        """
        Returns the lineage that is in common between two lineages

        Parameters
        ----------
        other : "Lineage"

        Returns
        -------
        Lineage
            Common lineage
        """
        for i, (s_lin, o_lin) in enumerate(zip(self, other)):
            if s_lin != o_lin:
                return Lineage(*self[:i])
        return Lineage(*self._fields)

    @property
    def name(self) -> Tuple[str, str]:
        """
        Get the lowest populated level and name of the taxon

        Returns
        -------
        Tuple[str, str]
            Tuple containing (level, name)
        """
        fields = self._fields
        for field in reversed(fields):
            ind = fields.index(field)
            name = self[ind]
            if name != "":
                return field, name
        return "Kingdom", "unclassified"

    @classmethod
    def from_str(cls, lineage_str: str, style: str = "gg") -> "Lineage":
        """
        Create `Lineage` instance from a lineage string

        Parameters
        ----------
        lineage_str : str
            Lineage in the form of a string
        style : {'gg', 'silva'}, optional
            The style of the lineage string
            Default is 'gg'

        Returns
        -------
        Lineage
            Instance of the `Lineage` class
        """
        if style == "gg":
            if lineage_str.startswith("k"):
                tax_list = lineage_str.split(";")
            elif lineage_str.startswith("p"):
                tax_list = ["Bacteria"] + lineage_str.split(";")
            else:
                raise ValueError("Incompatible lineage string")
        elif style == "silva":
            if lineage_str.startswith("D_0"):
                tax_list = lineage_str.split(";D_7")[0].split(";")
            elif lineage_str.startswith("D_1"):
                tax_list = ["Bacteria"] + lineage_str.split(";D_7")[0].split(";")
            else:
                raise ValueError("Incompatible lineage string")
        else:
            raise ValueError("Style has to be either 'gg' or 'silva'")
        taxa = [l.strip().rsplit("__", 1)[-1] for l in tax_list]
        return cls(*taxa)

    def to_str(self, style: str, level: str) -> str:
        """
        Return the string Lineage of the instance in requested 'style'

        Parameters
        ----------
        style : {'gg', 'silva'}
            The style of the lineage string
        level : str
            The lowest Lineage field that is to be populated

        Returns
        -------
        str
        """
        if level not in self._fields:
            raise ValueError(f"{level} not a valid field for Lineage")
        else:
            ind = self._fields.index(level)
            fields = self._fields[: ind + 1]
            data = self[: ind + 1]
        if style == "gg":
            prefix = [f.lower()[0] for f in fields]
        elif style == "silva":
            prefix = [f"D_{i}" for i in range(len(fields))]
        else:
            raise ValueError("Style needs to be either 'gg' or 'silva'")
        return ";".join(f"{p}__{v}" for p, v in zip(prefix, data))

    def __str__(self) -> str:
        """
        Get the lineage in the form of a string

        Returns
        -------
        str
            The lineage string in 'gg' format
        """
        return self.to_str(style="gg", level="Species")

    def to_dict(self, level: str) -> Dict[str, str]:
        """
        Get the lineage in the form of a dictionary

        Parameters
        ----------
        level : str
            The lowest Lineage field to be used to populate the dictionary
        """
        if level not in self._fields:
            raise ValueError(f"{level} not a valid field for Lineage")
        ind = self._fields.index(level)
        fields = self._fields[: ind + 1]
        return {field: tax for field, tax in zip(fields, self)}

    def get_superset(self, level: str) -> "Lineage":
        """
        Return a superset of the current lineage for the requested level

        Parameters
        ----------
        level : str
            The lowest Lineage field to be used to calculate the superset

        Returns
        -------
        Lineage
            Lineage instance that is a superset of current instance
        """
        if level not in self._fields:
            raise ValueError(f"{level} not a valid field for Lineage")
        ind = self._fields.index(level)
        tax = self[: ind + 1]
        return Lineage(*tax)

    @property
    def taxid(self) -> Tuple[str, int]:
        """
        Get the NCBI taxonomy id of the Lineage

        Returns
        -------
        Tuple[str, int]
            A tuple containing (taxonomy level, NCBI taxonomy id)
        """
        query = list(self)
        # species or subspecies level
        query.append(query[-2] + " " + query[-1].strip())
        # species level
        query[-2] = query[-3] + " " + query[-2].split(" ")[0].strip()
        taxid_dict = self._ncbi.get_name_translator(query)
        taxid_list = [12908]
        for taxa in reversed(query):
            if taxa != "" and taxa in taxid_dict:
                taxid_list = taxid_dict[taxa]
                break
        name = [q for q in reversed(query) if q != ""]
        if taxa != name[0] and taxa != name[1]:
            warning_msg = (
                f"Lowest level in {self} could not be queried. Using higher level"
            )
            LOG.logger.warning(warning_msg)
            warn(RuntimeWarning(warning_msg))
        if len(taxid_list) > 1:
            warning_msg = f"{self.name} has multiple taxids. Picking the first one"
            LOG.logger.warning(warning_msg)
            warn(RuntimeWarning(warning_msg))
        taxid = taxid_list[0]
        rank = self._fields[min(query.index(taxa), len(self._fields) - 1)]
        return rank, taxid

    @classmethod
    def from_taxid(cls, taxid: int) -> "Lineage":
        """
        Create `Lineage` instance from taxid

        Parameters
        ----------
        taxid : int
            A valid NCBI taxonomy id

        Returns
        -------
        "Lineage"
            Instance of the `Lineage` class
        """
        ncbi = NCBITaxa()
        lineage_taxids = ncbi.get_lineage(taxid)
        lineage_names = ncbi.get_taxid_translator(lineage_taxids)
        lineage_ranks = {
            v.capitalize(): k for k, v in ncbi.get_rank(lineage_taxids).items()
        }
        if "Superkingdom" in lineage_ranks:
            lineage_ranks["Kingdom"] = lineage_ranks["Superkingdom"]
            del lineage_ranks["Superkingdom"]
        taxa: Dict[str, str] = {}
        for field in cls._fields:
            if field in lineage_ranks:
                taxa[field] = lineage_names[lineage_ranks[field]]
            else:
                break
        return cls(**taxa)



"""
    Module that handles the execution of subprocesses and parsing of their outputs
"""

import subprocess
from typing import List, Optional

from ..logging import LOG


class Command:
    """
    Class that wraps functionality for running subprocesses and jobs on the cluster

    Parameters
    ----------
    cmd : str
        The command to be executed
    profile : {'local', 'sge'}
        The execution environment
    timeout : int, optional
        The time limit for the command
        If a process exceeds this time then it will be terminated
        Default is 1000

    Other Parameters
    ----------------
    project : str, optional
        The project under which to run the pipeline on the 'sge'
        Default value is None

    Attributes
    ----------
    cmd : str
        The command that will be executed.
        This includes the profile and resource specifics
    profile : {'local', 'sge'}
        The execution environment
    project : str
        The project under which to run the pipeline on the 'sge'
    output : str
        The 'stdout' of the command
    error : str
        The 'stderr' of the command
    status : str
        The status the the command
        One of {'success', 'failure', 'in progress', 'not started'}
    """

    _stdout: Optional[str] = None
    _stderr: Optional[str] = None
    process: Optional[subprocess.Popen] = None

    def __init__(self, cmd: str, profile: str, timeout: int = 1000, **kwargs) -> None:
        self.profile = profile
        project = kwargs.get("project")
        if profile == "sge":
            if project is None:
                raise ValueError("Project must be supplied if profile is sge")
        self.project = project if project else "None"
        self._cmd = self._build_cmd(cmd)
        self._timeout = timeout

    def _build_cmd(self, cmd: str) -> List[str]:
        """
        Builds the `cmd` for the current Command instance

        Parameters
        ----------
        cmd : str
            The command to be executed

        Returns
        -------
        str
            The final command to be executed
        """
        command: List[str] = []
        if self.profile == "local":
            pass
        elif self.profile == "sge":
            command.append("qsub")
            command.append("-P")
            command.append(self.project)
        else:
            raise ValueError("Unsupported profile! Choose either 'local' or 'sge'")
        command.extend(cmd.split(" "))
        return command

    def __str__(self) -> str:
        return self.cmd

    def __repr__(self) -> str:
        return f'<Command cmd="{self.cmd}" timeout={self._timeout}>'

    @property
    def cmd(self) -> str:
        """The command that will be executed"""
        return " ".join(self._cmd)

    def run(self, cwd: Optional[str] = None) -> subprocess.Popen:
        """
        Executes the command with the correct profile and resources

        Parameters
        ----------
        cwd : str, optional
            The directory in which the command is to be run
            Default is None which uses the current working directory

        Returns
        -------
        int
            The exit status of the command
        """
        # QUESTION: Replace this with asyncio.subprocess.create_subprocess_shell
        self.process = subprocess.Popen(
            self._cmd,
            cwd=cwd,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return self.process

    def wait(self) -> None:
        """Wait for the process to complete or terminate"""
        if self.process:
            stdout, stderr = self.process.communicate(timeout=self._timeout)
            self._stderr = stderr.decode("utf-8")
            self._stdout = stdout.decode("utf-8")

    def log(self) -> None:
        """Logs the stdout and stderr of the command execution to the log_file"""
        LOG.logger.info(f"Running command: {self.cmd}")
        LOG.logger.info("-" * 4 + " [STDOUT] " + "-" * 4)
        LOG.logger.success(self.output)
        LOG.logger.info("-" * 4 + " [STDERR] " + "-" * 4)
        LOG.logger.error(self.error)

    def proc_cmd_sync(self) -> bool:
        """
        Check whether the Command instance and subprocess.Popen process are in sync

        Returns
        -------
        bool
            True if both the `cmd` and `process` are the same
        """
        if self._cmd == self.process.args:
            return True
        else:
            return False

    @property
    def output(self) -> str:
        """Returns the output generated during execution of the command"""
        if self._stdout is not None:
            stdout = self._stdout
        else:
            if self.process:
                stdout, stderr = self.process.communicate(timeout=self._timeout)
                self._stdout = stdout.decode("utf-8")
                self._stderr = stderr.decode("utf-8")
            else:
                raise NotImplementedError(
                    "Please run the command before requesting output!"
                )
        return self._stdout

    @property
    def error(self) -> str:
        """Returns the error generated during execution of the command"""
        if self._stderr is not None:
            stderr = self._stderr
        else:
            if self.process:
                stdout, stderr = self.process.communicate(timeout=self._timeout)
                self._stdout = stdout.decode("utf-8")
                self._stderr = stderr.decode("utf-8")
            else:
                raise NotImplementedError(
                    "Please run the command before requesting errors!"
                )
        return self._stderr

    def update(self, cmd: str) -> None:
        """
        Update the `cmd` of the current Command instance

        Parameters
        ----------
        cmd : str
            The new command to be executed
        """
        self._cmd = self._build_cmd(cmd)
        if self.process:
            if not self.proc_cmd_sync():
                LOG.logger.warning(
                    "New command differs from executed command. Clearing previous run"
                )
                self._stdout = None
                self._stderr = None
                self.process = None

    @property
    def status(self) -> str:
        """
        Return the status of the command execution

        Returns
        -------
        str
            One of {'success', 'failure', 'in progress', 'not started'}
        """
        if self.process:
            poll = self.process.poll()
            if poll is None:
                return "in progress"
            if poll == 0:
                return "success"
            if poll > 0:
                return "failure"
            raise RuntimeError("Proces status is undefined")
        else:
            return "not started"



"""
    Module that defines the `NetworkGroup` object and methods to read, write and manipulate it
"""

from collections import defaultdict, Counter
from collections.abc import Collection
import pathlib
from itertools import product
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
from scipy.stats import pearsonr, chi2
import simplejson

from .network import Network

DType = List[Dict[str, Any]]


class NetworkGroup(Collection):
    """
    Class that represents a group of network objects
    These network objects are intended to be visualized together

    Parameters
    ----------
    networks : List[Network]
        The collection of networks to be grouped
        key = context-id, value = Network
    id_field : str
        The field to use while combining nodes
        Default value is "taxid"

    Attributes
    ----------
    graph : Union[nx.MultiGraph, nx.MultiDiGraph]
        The networkx multi-graph representation of the network
    nodes: DType
        The list of nodes in the network group
    links: DType
        The list of links in the network group
    contexts: DType
        The list of all contexts in the network group
    """

    def __init__(self, networks: List[Network], id_field: str = "taxid") -> None:
        self.id_field = id_field
        # dict(cid => dict(id_old => id_new))
        self.nodeid_map: Dict[int, Dict[str, str]] = dict()
        # dict(s_new-t_new => List[Tuple[cid, s_old-t_old], ...])
        self.linkid_revmap: Dict[str, List[Tuple[int, str]]] = defaultdict(list)
        self._networks = tuple(networks)
        if not networks or [n for n in networks if not isinstance(n, Network)]:
            raise ValueError(
                "The networks parameter must be a list of one or more networks"
            )
        self.graph = self._combine_networks(networks)

    def __contains__(self, key) -> bool:
        if key in range(len(self)):
            return True
        return False

    def __len__(self) -> int:
        return len(self._networks)

    def __iter__(self) -> Iterator:
        return iter(self._networks)

    def __repr__(self) -> str:
        n_nodes = len(self.nodes)
        n_links = len(self.links)
        n_contexts = len(self.contexts)
        return f"<NetworkGroup contexts={n_contexts} nodes={n_nodes} links={n_links}>"

    def __add__(self, other: "NetworkGroup") -> "NetworkGroup":
        """Combine two `NetworkGroup` objects and return a new `NetworkGroup` object
        The new `NetworkGroup` contains nodes and edges from both the input objects
        """
        if self.id_field != other.id_field:
            raise ValueError("Cannot add two NetworkGroups with different id_fields")
        networks = [*self._networks, *other._networks]
        return NetworkGroup(networks, id_field=self.id_field)

    def _combine_nodes(self, all_nodes: Dict[int, DType]) -> DType:
        """Combine nodes of individual networks into a single list"""
        nodes: DType = []
        node_hash: Dict[int, int] = dict()  # taxid => nodes.index
        if len(all_nodes) == 1:
            return all_nodes[0]
        for cid, network_nodes in all_nodes.items():
            self.nodeid_map[cid] = dict()
            for network_node in network_nodes:
                if network_node[self.id_field] not in node_hash:
                    id_ = len(nodes)
                    id_old = network_node["id"]
                    id_new = f"id{id_}"
                    nodes.append(
                        {
                            **network_node,
                            **{"id": id_new, "children": [], "abundance": None},
                        }
                    )
                    node_hash[network_node[self.id_field]] = id_
                    self.nodeid_map[cid][id_old] = id_new
                else:
                    id_old = network_node["id"]
                    ind = node_hash[network_node[self.id_field]]
                    id_new = nodes[ind]["id"]
                    self.nodeid_map[cid][id_old] = id_new
        return nodes

    def _combine_links(
        self, all_links: Dict[int, DType], inplace: bool = True
    ) -> DType:
        """Combine links of individual networks into a single list"""
        links = []
        if len(all_links) == 1:
            for link in all_links[0]:
                source, target = link["source"], link["target"]
                if inplace:
                    self.linkid_revmap[f"{source}-{target}"].append(
                        (0, f"{source}-{target}")
                    )
                links.append({**link, "context_index": 0})
            return links
        for cid, network_links in all_links.items():
            for link in network_links:
                source, target = link["source"], link["target"]
                new_source = self.nodeid_map[cid][source]
                new_target = self.nodeid_map[cid][target]
                if inplace:
                    self.linkid_revmap[f"{new_source}-{new_target}"].append(
                        (cid, f"{source}-{target}")
                    )
                links.append(
                    {
                        **link,
                        **{
                            "source": new_source,
                            "target": new_target,
                            "context_index": cid,
                        },
                    }
                )
        return links

    def _combine_networks(
        self, networks: List[Network]
    ) -> Union[nx.MultiGraph, nx.MultiDiGraph]:
        """
        Combine networks into a network group

        Parameters
        ----------
        networks : List[Network]
            The list of networks to be grouped

        Returns
        -------
        Union[nx.MultiGraph, nx.MultiDiGraph]
            The networkx graph of the network
        """
        nodes_dict = dict()
        links_dict = dict()
        contexts = []
        for cid, network in enumerate(networks):
            nodes_dict[cid] = network.nodes
            links_dict[cid] = network.links
            contexts.append(network.metadata)
        merged_nodes = self._combine_nodes(nodes_dict)
        merged_links = self._combine_links(links_dict)
        if all([n.graph.is_directed() for n in networks]):
            graph = nx.MultiDiGraph(contexts=contexts)
        else:
            graph = nx.MultiGraph(contexts=contexts)
        for node in merged_nodes:
            graph.add_node(node["id"], **node)
        for link in merged_links:
            graph.add_edge(link["source"], link["target"], **link)
        return graph

    @property
    def nodes(self) -> DType:
        """The list of nodes in the `NetworkGroup` and their corresponding properties"""
        return [data for _, data in self.graph.nodes(data=True)]

    @property
    def links(self) -> DType:
        """The list of links in the `NetworkGroup` and their corresponding properties"""
        return [data for _, _, data in self.graph.edges(data=True)]

    @property
    def contexts(self) -> DType:
        """The contexts for the group of networks"""
        return self.graph.graph["contexts"]

    def get_adjacency_vectors(self, key: str) -> pd.DataFrame:
        """
        Returns the adjacency matrix for each context as a `pd.DataFrame`

        Parameters
        ----------
        key : str
            The `edge` property to be used to contruct the vectors

        Returns
        -------
        pd.DataFrame:
            The DataFrame containing adjacency vectors as columns
        """
        ids = list(self.graph.nodes)
        size = len(ids) * len(ids)
        # NOTE: This will consider id1-id2 and id2-id1 as different (even for undirected)
        index = [f"{id1}-{id2}" for id1, id2 in product(ids, repeat=2)]
        n_contexts = len(self)
        adj_vector_df: pd.DataFrame = pd.concat(
            [
                pd.Series(np.zeros((size), dtype=float), index=index)
                for _ in range(n_contexts)
            ],
            join="outer",
            axis=1,
        )
        graph = self.graph
        # NOTE: networkx automatically handles directionality (source -> target) here
        for source, target, data in graph.edges(data=True, keys=False):
            cid = data["context_index"]
            id_ = f"{source}-{target}"
            adj_vector_df.loc[id_, cid] = data[key]
        return adj_vector_df

    def update_thresholds(
        self, interaction_threshold: float = 0.3, pvalue_threshold: float = 0.05
    ) -> None:
        """Update the thresholds on the networks

        Parameters
        ----------
        interaction_threshold : float, optional
            The value to which the interactions (absolute value) are to be thresholded
            To disable thresholding based on interaction value then pass in 0.0
            Default value is 0.3
        pvalue_threshold : float, optional
            This is the `alpha` value for pvalue cutoff
            Default value is 0.05
        """
        for context in self.graph.graph["contexts"]:
            context["interaction_threshold"] = interaction_threshold
            context["pvalue_threshold"] = pvalue_threshold
        for network in self._networks:
            network.interaction_threshold = interaction_threshold
            network.pvalue_threshold = pvalue_threshold

    def _filter_links(self, pvalue_filter: bool, interaction_filter: bool) -> DType:
        """
        The links of the networks after applying filtering

        Parameters
        ----------
        pvalue_filter : bool
            If True will use `pvalue_threshold` for filtering
        interaction_filter : bool
            If True will use `interaction_threshold` for filtering

        Returns
        -------
        DType
            The list of links in the network after applying thresholds
        """
        filtered_links_dict = dict()
        for cid, network in enumerate(self._networks):
            filtered_links_dict[cid] = network._filter_links(
                pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
            )
        merged_filtered_links = self._combine_links(filtered_links_dict, inplace=False)
        return merged_filtered_links

    def filter(self, pvalue_filter: bool, interaction_filter: bool) -> "NetworkGroup":
        """Filter network using pvalue and interaction thresholds

        Parameters
        ----------
        pvalue_filter : bool
            If `True` will use `pvalue_threshold` for filtering
        interaction_filter : bool
            If `True` will use `interaction_threshold` for filtering

        Returns
        -------
        "NetworkGroup"
            The filtered `NetworkGroup` object
        """
        nodes = {"nodes": self.nodes}
        links = {
            "links": self._filter_links(
                pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
            )
        }
        contexts = {"contexts": self.contexts}
        network_data = {**contexts, **nodes, **links}
        new_network = NetworkGroup.load_json(
            raw_data=network_data, id_field=self.id_field
        )
        return new_network

    def json(
        self, pvalue_filter: bool = False, interaction_filter: bool = False
    ) -> str:
        """
        Returns the network as a `JSON` string

        Parameters
        ----------
        pvalue_filter : bool
            If True will use `pvalue_threshold` for filtering
            Default  value is False
        interaction_filter : bool
            If True will use `interaction_threshold` for filtering
            Default  value is False

        Returns
        -------
        str
            The `JSON` string representation of the network
        """
        nodes = self.nodes
        links = self._filter_links(
            pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
        )
        contexts = self.contexts
        network = {"contexts": contexts, "nodes": nodes, "links": links}
        return simplejson.dumps(network, indent=2, sort_keys=True, ignore_nan=True)

    def write(
        self,
        fpath: str,
        pvalue_filter: bool = False,
        interaction_filter: bool = False,
        split_files: bool = False,
    ) -> None:
        """
        Write network to file as JSON

        Parameters
        ----------
        fpath : str
            The path to the `JSON` file
        pvalue_filter : bool
            If True will use `pvalue_threshold` for filtering
            Default  value is False
        interaction_filter : bool
            If True will use `interaction_threshold` for filtering
            Default  value is False
        split_files : bool
            If True will write networks into separate files
            Default value is False
        """
        if not split_files:
            with open(fpath, "w") as fid:
                fid.write(
                    self.json(
                        pvalue_filter=pvalue_filter,
                        interaction_filter=interaction_filter,
                    )
                )
        else:
            for cid, network in enumerate(self._networks):
                path = pathlib.Path(fpath)
                fname = f"{path.parent}/{cid}_{path.stem}{path.suffix}"
                network.write(
                    fname,
                    pvalue_filter=pvalue_filter,
                    interaction_filter=interaction_filter,
                )

    def to_network(self, method: str = "mean") -> Network:
        if method == "mean":
            method_func = np.mean
        else:
            raise NotImplemented("Only mean is supported right now")
        # Step1: Converge the metadata and cmetadata
        if isinstance(self.graph, nx.MultiDiGraph):
            GraphConstructor = nx.DiGraph
            directionality = "directed"
        elif isinstance(self.graph, nx.MultiGraph):
            directionality = "undirected"
            GraphConstructor = nx.Graph
        else:
            raise ValueError("Unknown graph type")
        emetadata = dict()
        cmetadata = dict()
        metadata_all = dict()
        for context in self.contexts:
            emetadata.update(context["experimental_metadata"])
            cmetadata.update(context["computational_metadata"])
            metadata_all.update(context)
        metadata = {
            **metadata_all,
            "experimental_metadata": emetadata,
            "computational_metadata": cmetadata,
            "interaction_type": "simple",
            "directionality": directionality,
        }
        networkgroup_graph = self.graph
        network_graph = GraphConstructor(**metadata)
        # Step2: Converge the nodes and links
        network_graph.add_nodes_from(networkgroup_graph.nodes(data=True))
        network_graph.remove_nodes_from(list(nx.isolates(networkgroup_graph)))
        for source, target, data in networkgroup_graph.edges(data=True, keys=False):
            if network_graph.has_edge(source, target):
                network_graph[source][target]["weight"].append(data.get("weight", 0.0))
                network_graph[source][target]["pvalue"].append(data.get("pvalue", 1.0))
            else:
                weight = data.get("weight", 0.0)
                pvalue = data.get("pvalue", 1.0)
                network_graph.add_edge(source, target, weight=[weight], pvalue=[pvalue])
        for source, target in network_graph.edges:
            weight = method_func(network_graph[source][target]["weight"])
            pvalue = method_func(network_graph[source][target]["pvalue"])
            network_graph[source][target]["weight"] = weight
            network_graph[source][target]["pvalue"] = pvalue
        return Network.load_graph(network_graph)

    @classmethod
    def load_json(
        cls,
        fpath: Optional[str] = None,
        raw_data: Optional[dict] = None,
        id_field: str = "taxid",
    ) -> "NetworkGroup":
        """
        Create a `NetworkGroup` object from network `JSON` file
        Either fpath or raw_data must be specified

        Parameters
        ----------
        fpath : str, optional
            The path to the network `JSON` file
        raw_data : dict, optional
            The raw data stored in the network `JSON` file

        Returns
        -------
        NetworkGroup
            The instance of the `NetworkGroup` class
        """
        if not raw_data and not fpath:
            raise ValueError("Either fpath or raw_data must be specified")
        if not raw_data and fpath:
            with open(fpath, "r") as fid:
                data = simplejson.load(fid)
        else:
            data: dict = raw_data
        n_networks = len(data["contexts"])
        all_node_dict = {n["id"]: n for n in data["nodes"]}
        data_dict: Dict[int, dict] = {
            n: {"nodes": [], "links": [], "metadata": {}} for n in range(n_networks)
        }
        unique_node_dict: Dict[int, dict] = {n: set() for n in range(n_networks)}
        for cid in range(n_networks):
            data_dict[cid]["metadata"] = {**data["contexts"][cid]}
        for link in data["links"]:
            link_cid = link["context_index"]
            source = all_node_dict[link["source"]]
            source_name = link["source"]
            target = all_node_dict[link["target"]]
            target_name = link["target"]
            data_dict[link_cid]["links"].append(link)
            if source_name not in unique_node_dict[link_cid]:
                data_dict[link_cid]["nodes"].append(source)
                unique_node_dict[link_cid].add(source_name)
            if target_name not in unique_node_dict[link_cid]:
                data_dict[link_cid]["nodes"].append(target)
                unique_node_dict[link_cid].add(target_name)
        networks: List[Network] = []
        for cid in range(n_networks):
            metadata = data_dict[cid]["metadata"]
            nodes = data_dict[cid]["nodes"]
            links = data_dict[cid]["links"]
            network_data = {**metadata, "nodes": nodes, "links": links}
            networks.append(Network.load_json(raw_data=network_data))
        return cls(networks, id_field=id_field)

    def get_consensus_network(
        self,
        cids: Optional[List[int]] = None,
        method: str = "simple_voting",
        parameter: float = 0.0,
    ) -> "NetworkGroup":
        """
        Get consensus network for the network defined by the `cids`

        Parameters:
        -----------
        cids : Optional[List[int]]
            The list of context ids that are to be used in the merger
            Default is None
        method : str, {"simple_voting", "scaled_sum"}
            Default value is simple_voting
        parameter : float
            Default value is 0.0 (which is the union of all the links)

        Returns
        -------
        consensus_network
            The `NetworkGroup` that represents the consensus network
        """

        # Method 1: Simple voting method
        def simple_voting(weights: pd.DataFrame, parameter: float) -> List[str]:
            """Perform a simple voting consensus"""
            size = weights.shape[1]  # no. of networks
            num_req_edges = np.floor(parameter * size)
            weights_signed = weights.copy()
            weights_signed[weights_signed > 0] = 1
            weights_signed[weights_signed < 0] = -1
            weights_signed.astype(int, copy=False)
            num_actual_edges = weights_signed.apply(
                lambda row: Counter(row).most_common(1)[0][-1], axis=1
            )
            indices_removal = weights.index[num_actual_edges < num_req_edges]
            return list(indices_removal)

        # Method 2: Scaled sum method
        def scaled_sum(weights: pd.DataFrame, parameter: float) -> List[str]:
            """Peform a scaled sum consensus"""
            size = weights.shape[1]  # no. of networks
            weights_scaled = weights.apply(lambda x: x / (np.abs(x).max()))
            parameter_scaled = (size - 1) * parameter
            indices_removal = weights.index[
                np.abs(weights_scaled.sum(axis=1)) < parameter_scaled
            ]
            return list(indices_removal)

        # Step1: Filter by "cids" and make copies of graphs
        graphs = []
        if not cids:
            cids = list(range(len(self.contexts)))
        for cid, network in enumerate(self._networks):
            if cid in cids:
                graphs.append(network.graph.copy())
        weights: pd.DataFrame = self.get_adjacency_vectors("weight")[cids]
        # Filling with dummy values
        weights.fillna(0.0, inplace=True)  # dummy weights = 0

        # Step 2: Apply voting method to each multiedge
        # indices_removal has {new_id_source}-{new_id_target}
        if method == "simple_voting":
            indices_removal = simple_voting(weights, parameter)
        elif method == "scaled_sum":
            indices_removal = scaled_sum(weights, parameter)
        else:
            raise ValueError("Only methods supported are simple_voting and scaled_sum")

        # Step 3: Use indices_removal on the networks
        graph_dict = dict(enumerate(graphs))
        for ind in indices_removal:
            for cid, ind_old in self.linkid_revmap[ind]:
                source_old, target_old = ind_old.split("-")
                graph_dict[cid].remove_edge(source_old, target_old)
        new_networks = [Network.load_graph(graph) for graph in graph_dict.values()]

        # Step 4: Return NetworkGroup object
        return NetworkGroup(new_networks, id_field=self.id_field)

    def combine_pvalues(self, cids: List[int]) -> "NetworkGroup":
        """
        Combine pvalues of links in the `cids` using Brown's p-value merging method

        Parameters:
        -----------
        cids : List[int]
            The list of context ids that are to be used in the merger

        Returns
        -------
        merged_network
            The `NetworkGroup` that contains the merged pvalues
        """

        # Step 1: Obtain the pvalues and weights
        weight_df: pd.DataFrame = self.get_adjacency_vectors("weight")[cids]
        pvalue_df: pd.DataFrame = self.get_adjacency_vectors("pvalue")[cids]

        # Filling with dummy values
        weight_df.fillna(0.0, inplace=True)  # dummy weights = 0
        pvalue_df.fillna(1.0, inplace=True)  # dummy pvalues = 1
        eps = np.finfo(np.float).eps
        pvalue_df.replace(0.0, eps, inplace=True)  # to prevent log(0)

        # Step 2: Calculate the combined pvalues using Browns method
        # E[psi] = 2 * k
        k = pvalue_df.shape[1]
        expected_value = 2 * k
        # Var[psi] = 4*k + 2 * sum{i<j} (3.263 * corr_ij + 0.710 * corr_ij^2 + 0.027 * corr_ij^3)
        variance = 4 * k
        for i in range(1, k):
            for j in range(0, i - 1):
                x_i = weight_df.iloc[:, i].values
                x_j = weight_df.iloc[:, j].values
                corr_ij, _ = pearsonr(x_i, x_j)
                cov_ij_approx = (
                    3.263 * corr_ij + 0.710 * (corr_ij ** 2) + 0.027 * (corr_ij ** 3)
                )
                variance += 2 * cov_ij_approx
        # df = 2 * E[psi]^2 / var[psi]
        degrees_of_freedom = 2 * (expected_value ** 2) / variance
        # c = var[psi] / (2 * E[psi])
        correction_factor = variance / (2 * expected_value)
        link_ids = pvalue_df.index
        chi_square = pvalue_df.apply(
            lambda x: -2.0 * np.log(x).sum() / correction_factor, axis=1
        )
        pvalues_combined = pd.Series(
            data=chi2.sf(chi_square, df=degrees_of_freedom), index=link_ids
        )

        # Step 3: Create new networks
        graphs = []
        for cid, network in enumerate(self._networks):
            if cid in cids:
                graphs.append(network.graph.copy())
        graph_dict = dict(enumerate(graphs))
        for ind in pvalues_combined.index:
            for cid, ind_old in self.linkid_revmap[ind]:
                source_old, target_old = ind_old.split("-")
                graph_dict[cid].edges[source_old, target_old][
                    "pvalue"
                ] = pvalues_combined[ind]
        new_networks = [Network.load_graph(graph) for graph in graphs]

        # Step 4: Return NetworkGroup object
        return NetworkGroup(new_networks, id_field=self.id_field)



"""
    Module that defines the schema for a valid OTU table
"""

from biom import Table
import numpy as np
from schematics.exceptions import ValidationError
from schematics.types import BaseType


class HeaderType(BaseType):
    """
    DataType that describes the expected structure and format for the sample headers
    """

    def validate_header(self, value):
        """ Check whether the header is valid """
        if any(not isinstance(v, str) for v in value):
            raise ValidationError("Invalid header. All samples must be strings")


class IndexType(BaseType):
    """ DataType that describes the expected structure and format for the OTU indices """

    def validate_index_str(self, value):
        if any(not isinstance(v, str) for v in value):
            raise ValidationError("Invalid index. All indices must be strings")

    def validate_index_unique(self, value):
        if len(value) != len(set(value)):
            raise ValidationError("Invalid index. All indices must be unqiue")


class DataType(BaseType):
    """" DataType that describes the expected structure and format for abundance values """

    def __init__(self, norm, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.norm = norm

    def validate_data_npfloat(self, value):
        if not value.dtype == "float64":
            raise ValidationError("Invalid data. Abundances must be float64")

    def validate_data_range(self, value):
        df = value.to_dataframe(dense=True)
        if df.values.min() < 0:
            raise ValidationError("Invalid data. Abundances cannot be negative")
        if self.norm:
            if df.values.max() > 1 or df.values.min() < 0:
                raise ValidationError("Invalid data. Abundances are not normalized")
            if any(not np.isclose(v, 1.0) for v in df.values.sum(axis=0)):
                raise ValidationError("Invalid data. Abundances are not normalized")


class SamplemetaType(BaseType):
    """ DataType that describes the expected structure and format for the sample metadata """

    def validate_samplemeta_columns(self, value):
        if len(value.columns) < 1:
            raise ValidationError("Invalid columns in sample metdata")

    def validate_samplemeta_index(self, value):
        if len(value.index) != len(set(value.index)):
            raise ValidationError(
                "Invalid index in sample metadata. All indices must be unqiue"
            )

    def validate_structure(self, value):
        if any(not isinstance(v, str) for v in value.index):
            raise ValidationError("Invalid index. All indices must be strings")
        if value.index.str.startswith("#").any():
            raise ValidationError(
                "Invalid sample metadata structure. Possibly incorrect header"
            )


class ObsmetaType(BaseType):
    """ DataType that describes the expected structure and format for the observation metadata """

    _req_keys = ["Kingdom", "Phylum", "Class", "Order", "Family", "Genus", "Species"]
    _extra_keys = ["Confidence", "Abundance"]

    def validate_index(self, value):
        if any(not isinstance(v, str) for v in value.index):
            raise ValidationError("Invalid index. All indices must be strings")

    def validate_obsmeta_headers(self, value):
        for col in value.columns:
            if col not in self._req_keys and col not in self._extra_keys:
                raise ValidationError(
                    f"Invalid observation metadata. Unknown attribute {col} present"
                )
        # Check if keys are in order
        # i.e. if genus is present everything above that level is present
        for key in self._req_keys:
            if key not in value.columns:
                ind = self._req_keys.index(key)
                if len(value.columns) != ind:
                    raise ValidationError(
                        f"Invalid observation metadata. Required attribute {key} not present"
                    )
                else:
                    break

    def validate_obsmeta_data(self, value):
        if self._extra_keys[0] in value.columns:
            confidence = value[self._extra_keys[0]]
            if confidence.dtype == float:
                cond1 = 0 <= confidence.min() <= 1
                cond2 = 0 <= confidence.max() <= 1
                if not cond1 and not cond2:
                    raise ValidationError(
                        "Invalid observation metadata. "
                        f"{self._extra_keys[0]} must have a value between 0 and 1"
                    )
            else:
                raise ValidationError(
                    "Invalid observation metadata. "
                    f"{self._extra_keys[0]} column must be of type float"
                )
            df = value.drop(self._extra_keys[0], axis=1)
        else:
            df = value
        if self._extra_keys[1] in value.columns:
            abundance = value[self._extra_keys[1]]
            if abundance.dtype != float:
                raise ValidationError(
                    "Invalid observation metadata. "
                    f"{self._extra_keys[1]} column must be of type float"
                )
            df = value.drop(self._extra_keys[1], axis=1)
        else:
            df = value
        for level, data in df.items():
            filt_data = data[data != ""]
            if level == "Species":
                query = filt_data[
                    ~filt_data.str.contains(r"^[a-zA-Z0-9-._ ]+(?<! )$")
                ].any()
                # TODO: FIXME:
                # if query:
                #     raise ValidationError(
                #         "Invalid observation metadata. "
                #         f"Taxonomy names are not standard: {query} is not allowed in {level}"
                #     )
            elif level in ["Kingdom", "Phylum", "Class", "Order", "Family", "Genus"]:
                query = filt_data[
                    ~filt_data.str.contains(r"^[a-zA-Z0-9-._ ]+(?<! )$")
                ].any()
                # TODO: FIXME:
                # if query:
                #     raise ValidationError(
                #         "Invalid observation metadata. "
                #         f"Taxonomy names are not standard: {query} is not allowed in {level}"
                #     )


class BiomType(BaseType):
    """
    DataType that describes the expected structure and format for the `biom.Table`

    Parameters
    ----------
    norm : bool, optional
        True if abundances are normalized
        Default value is False
    """

    def __init__(self, norm=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.norm = norm

    def validate_istable(self, value):
        """ Check whether the object is a `biom.Table` """
        if not isinstance(value, Table):
            raise ValidationError("Object must be a `biom.Table` instance")

    def validate_samples(self, value):
        """ Check whether the samples (columns) of the Table are valid """
        header_type = HeaderType()
        header_type.validate(value.ids(axis="sample"))

    def validate_index(self, value):
        """ Check whether the indices in the Table are valid """
        index_type = IndexType()
        index_type.validate(value.ids(axis="observation"))

    def validate_data(self, value):
        """ Check whether the data in the Table is valid """
        data_type = DataType(self.norm)
        data_type.validate(value)

    def validate_sample_metadata(self, value):
        """ Check whether the sample metadata in the Table is valid """
        samplemeta_type = SamplemetaType()
        sample_metadata = value.metadata_to_dataframe("sample")
        samplemeta_type.validate(sample_metadata)

    def validate_obs_metadata(self, value):
        """ Check whether the observation metadata in the Table is valid """
        obsmeta_type = ObsmetaType()
        obs_metadata = value.metadata_to_dataframe("observation")
        obsmeta_type.validate(obs_metadata)



#!/usr/bin/env python3

import pathlib
from micone import Otu
from warnings import warn


def main(otu_file: str, axis: str, column: str):
    otu = Otu.load_data(otu_file)
    if axis != "sample":
        raise NotImplementedError("Axis other than sample are not currently supported")
    if not column:
        warn("No column supplied for splitting")
        split_results = [("", otu)]
    else:
        func = lambda id_, md: md[column]
        split_results = otu.partition(axis, func)
    return split_results


if __name__ == "__main__":
    OTU_FILE = "${otu_file}"
    AXIS = "${axis}"
    COLUMN = "${column}"
    ID_ = "${meta.id}"
    split_results = main(OTU_FILE, AXIS, COLUMN)
    folder = pathlib.Path("split")
    folder.mkdir(exist_ok=True)
    filter_fun = lambda values, id_, md: True if any(values > 0.0) else False
    for label, split_otu in split_results:
        split_otu.otu_data.del_metadata(keys=[COLUMN], axis=AXIS)
        split_otu.filter(func=filter_fun, axis="observation")
        if label == "":
            split_otu.write(ID_, fol_path=str(folder))
        elif label is None:
            raise ValueError(f"Unable to split on column {COLUMN}")
        else:
            split_otu.write(ID_ + f"-{label}", fol_path=str(folder))



#!/usr/bin/env python

import glob
import os
import shutil

import pandas as pd


def main(file_glob, filter_flag):
    files = glob.glob(file_glob)
    filtered_dir = "filtered"
    os.mkdir(filtered_dir)
    for file in files:
        df = pd.read_table(file, index_col=0)
        obs_reject_flag = (df.sum(axis=1) == 0).any()
        sample_reject_flag = (df.sum(axis=0) == 0).any()
        if not filter_flag:
            obs_reject_flag = False
            sample_reject_flag = False
        if obs_reject_flag or sample_reject_flag:
            print("Rejected ", file)
        else:
            new_file = filtered_dir + "/" + os.path.basename(file).split(".temp")[0]
            shutil.copy(file, new_file)


if __name__ == "__main__":
    FILE_GLOB = "*.boot.temp"
    FILTER_FLAG = True if "${filter_flag}" == "True" else False
    main(FILE_GLOB, FILTER_FLAG)



"""
    Module that contains the logger configuration
"""

import pathlib
import random
import string

from loguru import logger

TMP_FOL = pathlib.Path("/tmp/micone")
TMP_FOL.mkdir(exist_ok=True)


class Log:
    """
    Class that handles the logging
    Wrapper around `loguru.logger`

    Parameters
    ----------
    folder : pathlib.Path
        Path to the folder containing the log files
        The folder must already exist
    format_str : str
        Format string for the log file
        "timestamp | level | message"
    disable : bool
        Flag to disable logging at the start
        Default value is True

    Attributes
    ----------
    path : pathlib.Path
        The path to the log file
    config : dict
        The dictionary used to configure the `loguru.logger`
    logger : loguru.logger
        The main logging object
    """

    def __init__(
        self, folder: pathlib.Path, format_str: str, disable: bool = True
    ) -> None:
        fname = (
            "".join(random.choices(string.ascii_lowercase + string.digits, k=16))
            + ".log"
        )
        self._folder = folder
        self.path = folder / fname
        self.config = {"handlers": [{"sink": str(self.path), "format": format_str}]}
        logger.configure(**self.config)
        self.logger = logger
        if disable:
            self.logger.disable("micone")

    def enable(self) -> None:
        """ Enable logging """
        self.logger.enable("micone")

    def cleanup(self, nfiles: int = 20) -> None:
        """
        Delete old logs if number > nfiles

        Parameters
        ----------
        nfiles : int
            The threshold for number of log files beyond which the older ones are deleted
            Default value is 10
        """
        log_files = list(self._folder.glob("*.log"))
        log_files_srtd = sorted(
            log_files, key=lambda x: x.stat().st_mtime, reverse=True
        )
        log_files_remove = log_files_srtd[nfiles:]
        for file in log_files_remove:
            file.unlink()


LOG = Log(TMP_FOL, format_str="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}")



#!/usr/bin/env python3

import json

from micone import Network


def main(
    base_name: str,
    corr_file: str,
    meta_file: str,
    cmeta_file: str,
    obsmeta_file: str,
    children_file: str,
    interaction_threshold: float,
    interaction_type: str,
) -> None:
    network = Network.load_data(
        interaction_file=corr_file,
        meta_file=meta_file,
        cmeta_file=cmeta_file,
        obsmeta_file=obsmeta_file,
        children_file=children_file,
        interaction_type=interaction_type,
        interaction_threshold=interaction_threshold,
        pvalue_correction=None,
    )
    network.write(base_name + "_network.json")
    # network.write(
    #     base_name + "_network_thres.json", pvalue_filter=False, interaction_filter=True
    # )


def create_cmetadata():
    cmetadata = {
        "denoise_cluster": "${meta.denoise_cluster}",
        "chimera_checking": "${meta.chimera_checking}",
        "taxonomy_assignment": "${meta.taxonomy_assignment}",
        "taxonomy_database": "${meta.taxonomy_database}",
        "network_inference": "${meta.network_inference}",
    }
    with open("cmetadata.json", "w") as fid:
        json.dump(cmetadata, fid)


if __name__ == "__main__":
    BASE_NAME = "${meta.id}"
    CORR_FILE = "${corr_file}"
    META_FILE = "${metadata_file}"
    if not META_FILE:
        raise ValueError(
            "The metadata file needs to be supplied via network.config in make_network_without_pvalue"
        )
    OBSMETA_FILE = "${obs_metadata}"
    CHILDREN_FILE = "${children_map}"
    INTERACTION_THRESHOLD = float("${interaction_threshold}")
    create_cmetadata()
    INTERACTION_TYPE = (
        "cozine_weight" if "${meta.network_inference}" == "cozine" else "correlation"
    )
    CMETA_FILE = "cmetadata.json"
    main(
        BASE_NAME,
        CORR_FILE,
        META_FILE,
        CMETA_FILE,
        OBSMETA_FILE,
        CHILDREN_FILE,
        INTERACTION_THRESHOLD,
        INTERACTION_TYPE,
    )



"""
    Module that defines the schema for interactions, pvalues, networks and their metadata
"""

import numpy as np
import pandas as pd
from schematics.exceptions import ValidationError
from schematics.models import Model
from schematics.types import (
    BaseType,
    DateType,
    DictType,
    FloatType,
    IntType,
    ListType,
    ModelType,
    StringType,
    UnionType,
)


class InteractionmatrixType(BaseType):
    """
    DataType that describes the expected structure of an interaction matrix

    Parameters
    ----------
    symm : bool, optional
        True if interaction matrix is expected to be symmetric
        Default value is False
    """

    def __init__(self, symm=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.symm = symm

    def validate_isdataframe(self, value):
        """Check whether the object is a pandas DataFrame"""
        if not isinstance(value, pd.DataFrame):
            raise ValidationError(
                "Interaction matrix must be a `pd.DataFrame` instance"
            )

    def validate_headers(self, value):
        """Check whether the rows and columns are the same"""
        if len(value.index) != len(value.columns):
            raise ValidationError(
                "Interaction matrix must have same number of rows and columns"
            )
        if any(value.index != value.columns):
            raise ValidationError(
                "Row and column header of an interaction matrix should match"
            )

    def validate_symmetry(self, value):
        """Check whether the the interaction matrix is symmetric"""
        if self.symm:
            if value.shape[0] != value.shape[1]:
                raise ValidationError("Interaction matrix is not symmetric")
            if not np.allclose(value, value.T):
                raise ValidationError("Interaction matrix is not symmetric")

    def validate_data(self, value):
        if not value.values.dtype == float and not value.values.dtype == int:
            raise ValidationError("Invalid data. Interactions must be int or float")


class CorrelationmatrixType(InteractionmatrixType):
    """DataType that describes the expected structure of a correlation matrix"""

    def __init__(self, *args, **kwargs):
        super().__init__(symm=True, *args, **kwargs)

    def validate_data_range(self, value):
        if value.values.max() > 1 or value.values.min() < -1:
            raise ValidationError("Correlation matrix must be bound by -1 and 1")


class PvaluematrixType(InteractionmatrixType):
    """DataType that describes the expected structure of a pvalue matrix"""

    def validate_data_range(self, value):
        if value.values.max() > 1 or value.values.min() < 0:
            raise ValidationError("Pvalue matrix must be bound by 0 and 1")


class PublicationModel(Model):
    """Model that describes the expected structure of the publication input"""

    date = DateType(required=True)
    authors = ListType(DictType(StringType), required=True)
    pubmed_id = StringType(required=True)


class MetadataModel(Model):
    """Model that describes the expected structure of the network metadata input"""

    host = StringType(required=True)
    condition = StringType(required=True)
    location = StringType(required=True)
    experimental_metadata = DictType(StringType, required=True)
    publication = ModelType(PublicationModel, required=True)
    description = StringType(required=True)


class ChildrenmapType(BaseType):
    """DataType that describes the expected structure of the children map dictionary"""

    def validate_keys(self, value):
        for k in value.keys():
            if not isinstance(k, str):
                raise ValidationError("Children map must have string keys")

    def validate_values(self, value):
        for v in value.values():
            if not isinstance(v, list):
                raise ValidationError(
                    "Children map must have lists of strings as values"
                )
            for elem in v:
                if not isinstance(elem, str):
                    raise ValidationError(
                        "Children map must have lists of strings as values"
                    )


class NodeModel(Model):
    """Model that describes the structure of one node in the network"""

    id = StringType(min_length=2, required=True)
    lineage = StringType(required=True)
    name = StringType(required=True)
    taxid = IntType(required=True)
    taxlevel = StringType(
        regex=r"(Kingdom|Phylum|Class|Order|Family|Genus|Species)", required=True
    )
    abundance = FloatType()
    children = ListType(StringType, required=True)


class NodesModel(Model):
    """Model that describes the structure of the nodes in the network"""

    nodes = ListType(ModelType(NodeModel), required=True)


class LinkModel(Model):
    """Model that describes the structure of one link in the network"""

    pvalue = FloatType()
    weight = FloatType(required=True)
    source = StringType(min_length=2, required=True)
    target = StringType(min_length=2, required=True)


class LinksModel(Model):
    """Model that describes the structure of one link in the network"""

    links = ListType(ModelType(LinkModel), required=True)


class NetworkmetadataModel(MetadataModel):
    """Model that describes the expected structure of the network metadata"""

    computational_metadata = DictType(
        UnionType(types=(StringType, FloatType)), required=True
    )
    directionality = StringType(regex=r"(undirected|directed)", required=True)
    interaction_type = StringType()


class ElistType(BaseType):
    """DataType that describes the expected structure of an edge list"""

    def validate_headers_index(self, value):
        if "source" not in value.columns:
            raise ValidationError("source column must be present in the edge list")
        if "target" not in value.columns:
            raise ValidationError("target column must be present in the edge list")
        if "weight" not in value.columns:
            raise ValidationError("weight column must be present in the edge list")
        if len(value["source"]) == len(set(value["source"])):
            raise ValidationError("Duplicate entries in source column not allowed")
        if len(value["target"]) == len(set(value["target"])):
            raise ValidationError("Duplicate entries in target column not allowed")



#!/usr/bin/env python3

# Script that groups OTU data on different taxa levels

import json
from typing import Tuple

from micone import Otu


# Group the otu_data on all the tax_levels
def grp_otu_data(otu_data: Otu, tax_level: str) -> Tuple[Otu, dict]:
    child_otu, child_groups = otu_data.collapse_taxa(tax_level)
    return child_otu, child_groups


if __name__ == "__main__":
    TAX_LEVEL: str = "${tax_level}"  # ['Family', 'Genus', 'Species']
    OTU_FILE = "${otu_file}"  # "otu.biom"
    OTU_DATA = Otu.load_data(OTU_FILE)
    child_otu, child_groups = grp_otu_data(OTU_DATA, TAX_LEVEL)
    fname = "${new_meta.id}"
    child_otu.write(fname, file_type="biom")
    with open(fname + "_children.json", "w") as fid:
        json.dump(child_groups, fid, indent=2, sort_keys=True)



#!/usr/bin/env python3

import networkx as nx
import numpy as np
import pandas as pd


def main(otu_file, network_file, output_file):
    otu_table = pd.read_table(otu_file, index_col=0)
    graph = nx.read_gml(network_file)
    nodes = otu_table.index
    size = (len(nodes), len(nodes))
    interaction_table = pd.DataFrame(
        np.zeros(size, dtype=float), index=nodes, columns=nodes
    )
    directed = nx.is_directed(graph)
    for source, target, data in graph.edges(data=True):
        if (source in nodes) and (target in nodes):
            interaction_table.loc[source, target] = data["weight"]
            if not directed:
                interaction_table.loc[target, source] = data["weight"]
    interaction_table.to_csv(output_file, sep="\\t", index=True, float_format="%.4f")


if __name__ == "__main__":
    OTU_FILE = "${otu_file}"
    NETWORK_FILE = "${network_file}"
    OUTPUT_FILE = "${meta.id}_corr.tsv"
    main(OTU_FILE, NETWORK_FILE, OUTPUT_FILE)



"""Top-level package for Microbial Interaction Database Pipeline."""

__author__ = """Dileep Kishore"""
__email__ = "dkishore@bu.edu"
__version__ = "0.7.1"

from .main import Lineage, Network, NetworkGroup, Otu



#!/usr/bin/env python3

from micone import Otu


def main(
    otu_file: str,
    rm_sparse_samples: str,
    rm_sparse_obs: str,
    axis: str,
    count_thres: int,
    prevalence_thres: float,
    abundance_thres: float,
    obssum_thres: float,
) -> Otu:
    otu = Otu.load_data(otu_file)
    if rm_sparse_samples == "true":
        otu = otu.rm_sparse_samples(count_thres=count_thres)
    if rm_sparse_obs == "true":
        otu = otu.rm_sparse_obs(
            prevalence_thres=prevalence_thres,
            abundance_thres=abundance_thres,
            obssum_thres=obssum_thres,
        )
    else:
        n_samples = otu.otu_data.shape[1]
        otu = otu.rm_sparse_obs(
            prevalence_thres=2 / n_samples, abundance_thres=0.001, obssum_thres=10
        )
    if axis != "None":
        otu_norm = otu.normalize(axis=axis)
    else:
        otu_norm = otu
    return otu_norm


if __name__ == "__main__":
    OTU_FILE = "${otu_file}"
    AXIS = "${axis}"
    RM_SPARSE_SAMPLES = "${rm_sparse_samples}"
    COUNT_THRES = int("${count_thres}")
    RM_SPARSE_OBS = "${rm_sparse_obs}"
    PREVALENCE_THRES = float("${prevalence_thres}")
    ABUNDANCE_THRES = float("${abundance_thres}")
    OBSSUM_THRES = int("${obssum_thres}")
    norm_otu = main(
        OTU_FILE,
        RM_SPARSE_SAMPLES,
        RM_SPARSE_OBS,
        AXIS,
        COUNT_THRES,
        PREVALENCE_THRES,
        ABUNDANCE_THRES,
        OBSSUM_THRES,
    )
    norm_otu.write("${meta.id}" + "_normalized_filtered")



"""
    Module that deals with validation of the pipeline results
"""

import ast
from itertools import chain, groupby
import multiprocessing as mp
from operator import itemgetter
import pathlib
import re
import subprocess
from typing import Dict, List, Tuple

import pandas as pd
import simplejson

from .otu_validator import OtuValidator
from .network_schema import NodesModel, LinksModel, NetworkmetadataModel


def validate_pipeline(history_file: pathlib.Path) -> str:
    """Check if the pipeline execution was successful
    We assume that the last run pipeline is the target

    Parameters
    ----------
    history_file : pathlib.Path
        The path to the nextflow history file

    Returns
    -------
    str
        success or fail
    """
    history: pd.DataFrame = pd.read_table(history_file, header=None, sep="\t")
    status = history.iloc[-1, 3]
    if status == "ERR":
        return "fail"
    elif status == "OK":
        return "success"
    else:
        raise ValueError("Unknown status")


def process_trace(trace_file: pathlib.Path) -> Dict[str, List[str]]:
    """Process the nextflow trace file and return the processes that succeeded and failed

    Parameters
    ----------
    trace_file : pathlib.Path
        The path to the trace file

    Returns
    -------
    Dict[str, List[str]]
        A dictionary that lists the "success" processes and "fail" processes
    """
    trace: pd.DataFrame = pd.read_table(trace_file, index_col=0, sep="\t")
    trace_grouped = trace.groupby("status")
    trace_summary = {"success": [], "fail": []}
    for group_label, ids in trace_grouped.groups.items():
        if group_label in {"CACHED", "COMPLETED"}:
            trace_summary["success"].extend([trace.loc[id_, "name"] for id_ in ids])
        elif group_label in {"FAILED"}:
            trace_summary["fail"].extend([trace.loc[id_, "name"] for id_ in ids])
        else:
            raise ValueError(f"Unknown group label {group_label}")
    return trace_summary


def validate_biom_file(biom_file: pathlib.Path) -> Tuple[str, pathlib.Path]:
    """Validate biom file

    Parameters
    ----------
    biom_file : pathlib.Path
        Path to the biom file

    Returns
    -------
    Tuple[str, pathlib.Path]
        result, path to biom file
    """
    otu_validator = OtuValidator(dtype="biom", ext="biom")
    try:
        otu_validator.load_validate(biom_file)
        result = "success"
    except:
        result = "fail"
    return result, biom_file


def validate_biom_results(
    biom_files: List[pathlib.Path], ncpus: int = 1
) -> Dict[str, List[pathlib.Path]]:
    """Validate the biom files in the pipeline output
    The biom files must contain observation and sample metadata

    Parameters
    ----------
    biom_files : List[pathlib.Path]
        The list of biom files to validate
    ncpus : int, optional
        The number of cpus to use
        Default value is 1

    Returns
    -------
    Dict[str, List[str]]
        Dictionary containing the "success" files and "fail" files
    """
    args = biom_files
    with mp.Pool(processes=ncpus) as pool:
        results = pool.map(validate_biom_file, args)
    results_dict = {
        k: [v[-1] for v in g]
        for k, g in groupby(sorted(results, key=itemgetter(0)), key=itemgetter(0))
    }
    return results_dict


def validate_network_file(network_file: pathlib.Path) -> Tuple[str, pathlib.Path]:
    """Validate network file

    Parameters
    ----------
    network_file : pathlib.Path
        Path to the network file

    Returns
    -------
    Tuple[str, pathlib.Path]
        result, path to network file
    """
    with open(network_file, "r") as fid:
        data = simplejson.load(fid)
    # Validation
    nodes_model = NodesModel({"nodes": data["nodes"]}, strict=False)
    links_model = LinksModel({"links": data["links"]}, strict=False)
    non_meta_keys = ["nodes", "links"]
    metadata = {k: v for k, v in data.items() if k not in non_meta_keys}
    networkmetadata_model = NetworkmetadataModel(metadata, strict=False)
    try:
        nodes_model.validate()
        links_model.validate()
        networkmetadata_model.validate()
        result = "success"
    except:
        result = "fail"
    return result, network_file


def validate_network_results(
    network_files: List[pathlib.Path], ncpus: int = 1
) -> Dict[str, List[pathlib.Path]]:
    """Validate the network files in the pipeline output
    These must be `Network` objects not `NetworkGroup` objects

    Parameters
    ----------
    network_files : List[pathlib.Path]
        The list of network files to validate
    ncpus : int, optional
        The number of cpus to use
        Default value is 1

    Returns
    -------
    Dict[str, List[str]]
        Dictionary containing the "success" files and "fail" files
    """
    args = network_files
    with mp.Pool(processes=ncpus) as pool:
        results = pool.map(validate_network_file, args)
    results_dict = {
        k: [v[-1] for v in g]
        for k, g in groupby(sorted(results, key=itemgetter(0)), key=itemgetter(0))
    }
    return results_dict


def validate_expected_results(
    execution_dir: pathlib.Path, trace_file: pathlib.Path, output_dir: pathlib.Path
) -> Dict[str, List[str]]:
    """Compare expected results (from nextflow config) with produced results
    This needs to be run in the conda environment

    Parameters
    ----------
    execution_dir : pathlib.Path
        The path to the pipeline execution directory
    trace_file : pathlib.Path
        The path to the trace file
    output_dir : pathlib.Path
        The path to the output directory

    Returns
    -------
    Dict[str, List[str]]
        List of modules that "success" or "fail"
    """
    cmd = ["nextflow", "config", "-flat", execution_dir]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    modules = list(
        chain(
            *[
                ast.literal_eval(re.search(r"(\['.*?'\])", s).group(1))
                for s in stdout.decode("utf-8").split("\n")
                if "selection =" in s
            ]
        )
    )
    results = {"success": [], "fail": []}
    for module in modules:
        # Check the trace file for success
        trace = process_trace(trace_file)
        module_failures = [module for failure in trace["fail"] if module in failure]
        # Check the outputs dir for folder
        module_dirs = [dir for dir in output_dir.glob(f"**/{module}*") if dir.is_dir()]
        if not module_failures and module_dirs:
            results["success"].append(module)
        else:
            results["fail"].append(module)
    return results


def check_results(pipeline_dir: pathlib.Path, procs: int):
    """Run all the validation functions on the pipeline_dir
    1. validate_pipeline(history_file)
    2. process_trace(trace_file)
    3. validate_biom_results(biom_files, procs)
    4. validate_network_results(network_files, procs)
    5. validate_expected_results(pipeline_dir, trace_file, output_dir)

    Returns
    -------
    tuple
        A tuple containing results from the processes in the order specified above
    """
    history_file = pipeline_dir / ".nextflow/history"
    trace_file = pipeline_dir / "trace.txt"
    output_dir = pipeline_dir / "outputs"
    biom_files = list(pathlib.Path("outputs/tax_assignment").glob("**/*.biom"))
    network_files = list(
        pathlib.Path("outputs/network_inference/network").glob(
            "make_network_with_pvalue/**/*_network.json"
        )
    ) + list(
        pathlib.Path("outputs/network_inference/network").glob(
            "make_network_without_pvalue/**/*_network.json"
        )
    )
    return (
        validate_pipeline(history_file),
        process_trace(trace_file),
        validate_biom_results(biom_files, procs),
        validate_network_results(network_files, procs),
        validate_expected_results(pipeline_dir, trace_file, output_dir),
    )



"""
    Module that defines the `Network` object and methods to read, write and manipulate it
"""

from typing import Any, Dict, List, Optional, Tuple, Union
from warnings import warn

import networkx as nx
import numpy as np
import pandas as pd
import simplejson
from statsmodels.stats.multitest import multipletests

from . import Lineage
from ..validation import (
    InteractionmatrixType,
    CorrelationmatrixType,
    PvaluematrixType,
    ObsmetaType,
    MetadataModel,
    ChildrenmapType,
    NodesModel,
    LinksModel,
    NetworkmetadataModel,
    ElistType,
)


DType = List[Dict[str, Any]]
LinkDType = Tuple[str, str, Dict[str, float]]


class Network:
    """
    Class that represents a network object

    Parameters
    ----------
    nodes : List[str]
        The list of nodes in the network
    links : List[LinkDType]
        The list of links in the network
        Each link is a dict and must contain: 'source', 'target', 'weight', 'pvalue' as keys
    metadata : dict
        The metadata for the whole network (general and experiment)
        Must contain 'host', 'condition', 'location', 'experimental_metadata', 'pubmed_id',
        'description', 'date', 'authors
    cmetadata : dict
        The computational metadata for the whole network
        Must contain information as to how the network was generated
    obs_metadata : pd.DataFrame
        The `DataFrame` containing taxonomy information for the nodes of the network
        If this contains an 'Abundance' column then it is incorporated into the network
    children_map : dict, optional
        The dictionary that contains the mapping {obs_id => [children]}
    interaction_type : str, optional
        The type of interaction encoded by the edges of the network
        Default value is correlation
    interaction_threshold : float, optional
        The value to which the interactions (absolute value) are to be thresholded
        To disable thresholding based on interaction value then pass in 0.0
        Default value is 0.3
    pvalue_threshold : float, optional
        This is the `alpha` value for pvalue cutoff
        Default value is 0.05
    pvalue_correction : str, optional
        The method to use for multiple hypothesis correction
        Default value is 'fdr_bh'
        Set to None to turn off multiple hypothesis correction
        Use `Network.pcorr_methods` to get the list of supported methods
    directed : bool, optional
        True if network is directed
        Default value is False

    Attributes
    ----------
    graph : Union[nx.Graph, nx.DiGraph]
        The networkx graph representation of the network
    nodes : DType
        The list of nodes in the network and their corresponding properties
    links : DType
        The list of links in the network and their corresponding properties
    metadata : Dict[str, Any]
        The metadata for the network

    Examples
    --------
    >>> network = Network.load_data()
    """

    def __init__(
        self,
        nodes: List[str],
        links: List[LinkDType],
        metadata: dict,
        cmetadata: dict,
        obs_metadata: pd.DataFrame,
        children_map: Optional[dict] = None,
        interaction_type: str = "correlation",
        interaction_threshold: float = 0.3,
        pvalue_threshold: float = 0.05,
        pvalue_correction: Optional[str] = "fdr_bh",
        directed: bool = False,
    ) -> None:
        self.interaction_threshold = interaction_threshold
        self.pvalue_threshold = pvalue_threshold
        obsmeta_validator = ObsmetaType()
        obsmeta_validator.validate(obs_metadata)
        metadata_model = MetadataModel(metadata, strict=False)
        metadata_model.validate()
        if children_map:
            children_validator = ChildrenmapType()
            children_validator.validate(children_map)
        pvalues = np.array([link[2]["pvalue"] for link in links])
        if pvalue_correction:
            corrected_pvalues = self._correct_pvalues(
                pvalues, pvalue_correction, pvalue_threshold
            )
        else:
            corrected_pvalues = pvalues
        corrected_links: List[LinkDType] = []
        for i, link in enumerate(links):
            source, target = link[0], link[1]
            corrected_links.append(
                (source, target, {**link[2], "pvalue": corrected_pvalues[i]})
            )
        if "pvalue_threshold" not in cmetadata:
            cmetadata["pvalue_threshold"] = pvalue_threshold
        if "pvalue_correction" not in cmetadata:
            cmetadata["pvalue_correction"] = pvalue_correction
        if "interaction_threshold" not in cmetadata:
            cmetadata["interaction_threshold"] = interaction_threshold
        self.graph = self._create_graph(
            nodes,
            corrected_links,
            metadata,
            cmetadata,
            obs_metadata,
            children_map,
            interaction_type,
            directed,
        )
        nodes_model = NodesModel({"nodes": self.nodes}, strict=False)
        nodes_model.validate()
        links_model = LinksModel({"links": self.links}, strict=False)
        links_model.validate()
        networkmetadata_model = NetworkmetadataModel(self.metadata, strict=False)
        networkmetadata_model.validate()

    def __repr__(self) -> str:
        n_nodes = len(self.nodes)
        n_links = len(self.links)
        directionality = self.metadata["directionality"]
        interaction_type = self.metadata["interaction_type"]
        string = (
            f"<Network nodes={n_nodes} links={n_links} "
            f"type={interaction_type} directionality={directionality}>"
        )
        return string

    @staticmethod
    def _verify_integrity(interactions: pd.DataFrame, pvalues: pd.DataFrame) -> None:
        """
        Verify whether the interactions and pvalue matrices match

        Parameters
        ----------
        interactions : pd.DataFrame
            The `DataFrame` containing the interaction matrix
        pvalues : pd.DataFrame
            The `DataFrame` containing the pvalue matrix
        """
        if any(interactions.index != pvalues.index) or any(
            interactions.columns != pvalues.columns
        ):
            raise ValueError(
                "Interaction and pvalue matrices do not have matching indices"
            )

    @property
    def pcorr_methods(self) -> List[str]:
        """
        Returns list supported pvalue correction methods
        """
        methods = [
            "bonferroni",
            "sidak",
            "holm-sidak",
            "holm",
            "simes-hochberg",
            "hommel",
            "fdr_bh",
            "fdr_by",
            "fdr_tsbh",
            "fdr_tsbky",
        ]
        return methods

    def _correct_pvalues(
        self, pvalues: np.array, method: str, pvalue_threshold: float
    ) -> pd.DataFrame:
        """
        Correct pvalues using 'method'

        Parameters
        ----------
        pvalues : pd.DataFrame
            Raw uncorrected pvalues
        method : str
            Method to be used to correct the pvalues.
            Use `Network.pcorr_methods` to get the list of supported methods
        pvalue_threshold : float
            The value of alpha (FWER) to be used for the correction

        Returns
        -------
        pd.DataFrame
            DataFrame containing corrected pvalues
        """
        if method not in self.pcorr_methods:
            raise ValueError(
                f"Method {method} not supported. Must be one of {self.pcorr_methods}"
            )
        _, pvals_correct, *_ = multipletests(
            pvalues, alpha=pvalue_threshold, method=method
        )
        return pvals_correct

    @staticmethod
    def _create_graph(
        nodes: List[str],
        links: List[LinkDType],
        emetadata: dict,
        cmetadata: dict,
        obs_metadata: pd.DataFrame,
        children_map: Optional[dict],
        interaction_type: str,
        directed: bool,
    ) -> Union[nx.Graph, nx.DiGraph]:
        """
        Create network from interaction matrix, pvalue matrix, metadata dictionary,
        lineage table and children mapping

        Parameters
        ----------
        nodes : List[str]
            The list of nodes in the network
        links : List[LinkDType]
            The list of links in the network
        emetadata : dict
            The dictionary of general and experimental metadata
        cmetadata : dict
            The dictionary of computational metadata
        obs_metadata : pd.DataFrame
            The `DataFrame` containing taxonomy information for the nodes of the network
        children_map : dict
            The dictionary that contains the mapping {obs_id => [children]}
        interaction_type : str
            The type of interaction encoded by the edges of the network
        directed : bool
            Flag to determine whether the network is directed or not

        Returns
        -------
        Union[nx.Graph, nx.DiGraph]
            The networkx graph of the network
        """
        directionality = "directed" if directed else "undirected"
        metadata = {
            **emetadata,
            "computational_metadata": cmetadata,
            "interaction_type": interaction_type,
            "directionality": directionality,
        }
        if directed:
            graph = nx.DiGraph(**metadata)
        else:
            graph = nx.Graph(**metadata)
        abundance_flag = "Abundance" in obs_metadata.columns
        for node in nodes:
            if abundance_flag:
                lineage = Lineage(
                    **obs_metadata.drop("Abundance", axis=1).loc[node].to_dict()
                )
                abundance = obs_metadata.loc[node].Abundance
            else:
                if node not in obs_metadata.index:
                    lineage = Lineage(Kingdom="Bacteria")
                    warn(
                        UserWarning(
                            f"{node} not found in obs_metadata. Assigning lineage as Bacteria"
                        )
                    )
                else:
                    lineage = Lineage(**obs_metadata.loc[node].to_dict())
                abundance = None
            if children_map:
                children = children_map.get(node, [])
            else:
                children = []
            sup_lineage = lineage.get_superset(lineage.name[0])
            graph.add_node(
                node,
                id=node,
                lineage=sup_lineage.to_str(style="gg", level=sup_lineage.name[0]),
                name=sup_lineage.name[1],
                taxid=sup_lineage.taxid[1],
                taxlevel=sup_lineage.name[0],
                abundance=abundance,
                children=children,
            )
        for link in links:
            source, target = link[0], link[1]
            # NOTE: Self-loops are not allowed
            if source == target:
                continue
            weight, pvalue = link[2]["weight"], link[2]["pvalue"]
            graph.add_edge(
                source,
                target,
                source=source,
                target=target,
                weight=weight,
                pvalue=pvalue,
            )
        return graph

    @property
    def nodes(self) -> DType:
        """The list of nodes in the network and their corresponding properties"""
        return [data for _, data in self.graph.nodes(data=True)]

    @property
    def links(self) -> DType:
        """The list of links in the network and their corresponding properties"""
        return [data for _, _, data in self.graph.edges(data=True)]

    @property
    def metadata(self) -> Dict[str, Any]:
        """The metadata for the network"""
        return self.graph.graph

    def get_adjacency_table(self, key: str) -> pd.DataFrame:
        """
        Returns the adjacency table representation for the requested `key`
        This method does not support `Graph`

        Parameters
        ----------
        key : str
            The `edge` property to be used to construct the table

        Returns
        -------
        pd.DataFrame:
            The adjacency table
        """
        ids = list(self.graph.nodes)
        size = len(ids)
        adj_table = pd.DataFrame(
            data=np.zeros((size, size), dtype=float), index=ids, columns=ids
        )
        graph = self.graph
        for source_id, target_id, data in graph.edges(data=True):
            adj_table[source_id][target_id] = data[key]
            if not self.graph.is_directed():
                adj_table[target_id][source_id] = data[key]
        return adj_table

    def _filter_links(self, pvalue_filter: bool, interaction_filter: bool) -> DType:
        """
        The links of the network after applying filtering

        Parameters
        ----------
        pvalue_filter : bool
            If True will use `pvalue_threshold` for filtering
        interaction_filter : bool
            If True will use `interaction_threshold` for filtering

        Returns
        -------
        DType
            The list of links in the network after applying thresholds
        """
        interaction_threshold = abs(self.interaction_threshold)
        interaction_func = lambda x: abs(x["weight"]) >= interaction_threshold
        pvalues = [l["pvalue"] for l in self.links]
        no_pvalues = np.isnan(pvalues).all()
        if no_pvalues:
            pvalue_func = lambda _: True
        else:
            pvalue_func = lambda x: x["pvalue"] <= self.pvalue_threshold
        if pvalue_filter and interaction_filter:
            filter_func = lambda x: interaction_func(x) and pvalue_func(x)
        elif interaction_filter:
            filter_func = interaction_func
        elif pvalue_filter:
            filter_func = pvalue_func
        else:
            return self.links
        links_thres = list(filter(filter_func, self.links))
        return links_thres

    def filter(self, pvalue_filter: bool, interaction_filter: bool) -> "Network":
        """Filter network using pvalue and interaction thresholds

        Parameters
        ----------
        pvalue_filter : bool
            If `True` will use `pvalue_threshold` for filtering
        interaction_filter : bool
            If `True` will use `interaction_threshold` for filtering

        Returns
        -------
        "Network"
            The filtered `Network` object
        """
        nodes = {"nodes": self.nodes}
        links = {
            "links": self._filter_links(
                pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
            )
        }
        metadata = self.metadata
        network_data = {**metadata, **nodes, **links}
        new_network = Network.load_json(raw_data=network_data)
        return new_network

    @classmethod
    def load_data(
        cls,
        interaction_file: str,
        meta_file: str,
        cmeta_file: str,
        obsmeta_file: str,
        pvalue_file: Optional[str] = None,
        children_file: Optional[str] = None,
        interaction_type: str = "correlation",
        interaction_threshold: float = 0.3,
        pvalue_threshold: float = 0.05,
        pvalue_correction: Optional[str] = "fdr_bh",
        directed: bool = False,
    ) -> "Network":
        """
        Create a `Network` object from files (interaction tables and other metadata)

        Parameters
        ----------
        interaction_file : str
            The `tsv` file containing the matrix of interactions
        meta_file : str
            The `json` file containing the metadata for the whole network (general and experiment)
        cmeta_file : str
            The `json` file containings the computational metadata for the whole network
        obsmeta_file : str
            The `csv` file containing taxonomy information for the nodes of the network
        pvalue_file : str, optional
            The `tsv` file containing the matrix of pvalues
            Default is None
        children_file : str, optional
            The `json` file containing the mapping between observations and their children
        interaction_type : str, optional
            The type of interaction encoded by the edges of the network
            Default value is correlation
        interaction_threshold : float, optional
            The value to which the interactions (absolute value) are to be thresholded
            To disable thresholding based on interaction value then pass in 0.0
            Default value is 0.3
        pvalue_threshold : float, optional
            This is the `alpha` value for pvalue cutoff
            Default value is 0.05
        pvalue_correction : str, optional
            The method to use for multiple hypothesis correction
            Default value is 'fdr_bh'
            Set to None to turn off multiple hypothesis correction
            Use `Network.pcorr_methods` to get the list of supported methods
        directed : bool
            True if network is directed
            Default value is False

        Returns
        -------
        Network
            The instance of the `Network` class
        """
        # Load and validate interaction matrix
        interactions = pd.read_table(interaction_file, index_col=0)
        if interaction_type == "correlation":
            interaction_validator = CorrelationmatrixType()
        else:
            interaction_validator = InteractionmatrixType(symm=not directed)
        interaction_validator.validate(interactions)
        # Load and validate pvalue matrix
        if pvalue_file is not None:
            pvalues = pd.read_table(pvalue_file, index_col=0)
            cls._verify_integrity(interactions, pvalues)
            pvalue_validator = PvaluematrixType(symm=directed)
            pvalue_validator.validate(pvalues)
        else:
            pvalues = None
        # If undirected convert to upper triangular matrix
        if directed:
            interaction_mat = interactions.values
        else:
            interaction_mat = np.triu(interactions.values)
        row_inds, col_inds = interaction_mat.nonzero()
        # Calculate nodes and links
        nodes = list(interactions.index)
        links: List[LinkDType] = []
        for rind, cind in zip(row_inds, col_inds):
            links.append(
                (
                    interactions.index[rind],
                    interactions.columns[cind],
                    {
                        "weight": interactions.iloc[rind, cind],
                        "pvalue": pvalues.iloc[rind, cind] if pvalue_file else np.nan,
                    },
                )
            )
        # Load metadata
        with open(meta_file, "r") as fid:
            metadata = simplejson.load(fid)
        with open(cmeta_file, "r") as fid:
            cmetadata = simplejson.load(fid)
        if pvalue_file:
            extra_compdata = {
                "interaction_threshold": interaction_threshold,
                "pvalue_threshold": pvalue_threshold,
                "pvalue_correction": pvalue_correction,
            }
        else:
            extra_compdata = {"interaction_threshold": interaction_threshold}
        cmetadata = {**cmetadata, **extra_compdata}
        obs_metadata = pd.read_csv(obsmeta_file, index_col=0, na_filter=False)
        if children_file is not None:
            with open(children_file, "r") as fid:
                children_map = simplejson.load(fid)
        else:
            children_map = None
        network = cls(
            nodes,
            links,
            metadata,
            cmetadata,
            obs_metadata,
            children_map,
            interaction_type,
            interaction_threshold,
            pvalue_threshold,
            pvalue_correction,
            directed,
        )
        return network

    def json(
        self, pvalue_filter: bool = False, interaction_filter: bool = False
    ) -> str:
        """
        Returns the network as a `JSON` string

        Parameters
        ----------
        pvalue_filter : bool
            If True will use `pvalue_threshold` for filtering
            Default  value is False
        interaction_filter : bool
            If True will use `interaction_threshold` for filtering
            Default  value is False

        Returns
        -------
        str
            The `JSON` string representation of the network
        """
        nodes = {"nodes": self.nodes}
        links = {
            "links": self._filter_links(
                pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
            )
        }
        metadata = self.metadata
        network = {**metadata, **nodes, **links}
        return simplejson.dumps(network, indent=2, sort_keys=True, ignore_nan=True)

    def write(
        self, fpath: str, pvalue_filter: bool = False, interaction_filter: bool = False
    ) -> None:
        """
        Write network to file as JSON

        Parameters
        ----------
        fpath : str
            The path to the `JSON` file
        pvalue_filter : bool
            If True will use `pvalue_threshold` for filtering
            Default  value is False
        interaction_filter : bool
            If True will use `interaction_threshold` for filtering
            Default  value is False
        """
        with open(fpath, "w") as fid:
            fid.write(
                self.json(
                    pvalue_filter=pvalue_filter, interaction_filter=interaction_filter
                )
            )

    @classmethod
    def load_json(
        cls, fpath: Optional[str] = None, raw_data: Optional[dict] = None
    ) -> "Network":
        """
        Create a `Network` object from a network `JSON` file
        Either fpath or raw_data must be specified

        Parameters
        ----------
        fpath : str, optional
            The path to the network `JSON` file
        raw_data : dict, optional
            The raw data stored in the network `JSON` file

        Returns
        -------
        Network
            The instance of the `Network` class
        """
        if not raw_data and not fpath:
            raise ValueError("Either fpath or raw_data must be specified")
        if not raw_data and fpath:
            with open(fpath, "r") as fid:
                data = simplejson.load(fid)
        else:
            data = raw_data
        # Validation
        nodes_model = NodesModel({"nodes": data["nodes"]}, strict=False)
        nodes_model.validate()
        links_model = LinksModel({"links": data["links"]}, strict=False)
        links_model.validate()
        non_meta_keys = ["nodes", "links"]
        metadata = {k: v for k, v in data.items() if k not in non_meta_keys}
        networkmetadata_model = NetworkmetadataModel(metadata, strict=False)
        networkmetadata_model.validate()
        # Variable assignment
        cmetadata = data["computational_metadata"]
        interaction_type = data["interaction_type"]
        interaction_threshold = cmetadata["interaction_threshold"]
        pvalue_threshold = cmetadata["pvalue_threshold"]
        pvalue_correction = None
        directed = data["directionality"] == "directed"
        nodes: List[str] = []
        links: List[LinkDType] = []
        lineages: List[dict] = []
        children_map: Dict[str, List[str]] = {}
        for node in data["nodes"]:
            nodes.append(node["id"])
            lineage = Lineage.from_str(node["lineage"]).to_dict("Species")
            children_map[node["id"]] = node["children"]
            abundance = node.get("abundance", np.nan)
            if abundance is not None:
                lineages.append({**lineage, **dict(Abundance=abundance)})
            else:
                lineages.append(lineage)
        obs_metadata = pd.DataFrame(lineages, index=nodes)
        for link in data["links"]:
            source, target = link["source"], link["target"]
            if link["weight"] is None:
                weight = np.nan
            else:
                weight = link["weight"]
            if link["pvalue"] is None:
                pvalue = np.nan
            else:
                pvalue = link["pvalue"]
            links.append((source, target, {"weight": weight, "pvalue": pvalue}))
        network = cls(
            nodes,
            links,
            metadata,
            cmetadata,
            obs_metadata,
            children_map,
            interaction_type,
            interaction_threshold,
            pvalue_threshold,
            pvalue_correction,
            directed,
        )
        return network

    @classmethod
    def load_elist(
        cls,
        elist_file: str,
        meta_file: str,
        cmeta_file: str,
        obsmeta_file: str,
        children_file: Optional[str] = None,
        interaction_type: str = "correlation",
        interaction_threshold: float = 0.3,
        pvalue_threshold: float = 0.05,
        pvalue_correction: Optional[str] = "fdr_bh",
        directed: bool = False,
    ) -> "Network":
        """
        Create `Network` instance from an edge list and associated metadata

        Parameters
        ----------
        elist_file : str
            The csv file containing the list of edges and their associated metadata
        meta_file : dict
            The file containing metadata for the whole network (general and experiment)
            Must contain 'host', 'condition', 'location', 'experimental_metadata', 'pubmed_id',
            'description', 'date', 'authors'
        cmeta_file : dict
            The computational metadata for the whole network
            Must contain information as to how the network was generated
        obsmeta_file : str
            The csv file contanining taxonomy information for the nodes of the network
            If this contains an 'Abundance' column then it is incorporated into the network
        children_file : str, optional
            The json file that describes the mapping between {obs_id => [children]}
        interaction_type : str, optional
            The type of interaction encoded by the edges of the network
            Default value is correlation
        interaction_threshold : float, optional
            The value to which the interactions (absolute value) are to be thresholded
            To disable thresholding based on interaction value then pass in 0.0
            Default value is 0.3
        pvalue_threshold : float, optional
            This is the `alpha` value for pvalue cutoff
            Default value is 0.05
        pvalue_correction : str, optional
            The method to use for multiple hypothesis correction
            Default value is 'fdr_bh'
            Set to None to turn off multiple hypothesis correction
        directed : bool, optional
            True if network is directed
            Default value is False

        Returns
        -------
        Network
            The instance of the `Network` class
        """
        elist: pd.DataFrame = pd.read_csv(elist_file, na_filter=False)
        elist_type = ElistType()
        elist_type.validate(elist)
        nodes = list(set([*elist["source"], *elist["target"]]))
        links: List[LinkDType] = []
        pvalue_flag = "pvalue" in elist.columns
        for entry in elist.to_dict("records"):
            source, target, weight = entry["source"], entry["target"], entry["weight"]
            links.append(
                (
                    source,
                    target,
                    {
                        "weight": weight,
                        "pvalue": entry["pvalue"] if pvalue_flag else np.nan,
                    },
                )
            )
        with open(meta_file, "r") as fid:
            metadata = simplejson.load(fid)
        with open(cmeta_file, "r") as fid:
            cmetadata = simplejson.load(fid)
        if pvalue_flag:
            extra_compdata = {
                "interaction_threshold": interaction_threshold,
                "pvalue_threshold": pvalue_threshold,
                "pvalue_correction": pvalue_correction,
            }
        else:
            extra_compdata = {"interaction_threshold": interaction_threshold}
        cmetadata = {**cmetadata, **extra_compdata}
        obs_metadata = pd.read_csv(obsmeta_file, index_col=0, na_filter=False)
        if children_file is not None:
            with open(children_file, "r") as fid:
                children_map = simplejson.load(fid)
        else:
            children_map = None
        network = cls(
            nodes,
            links,
            metadata,
            cmetadata,
            obs_metadata,
            children_map,
            interaction_type,
            interaction_threshold,
            pvalue_threshold,
            pvalue_correction,
            directed,
        )
        return network

    @classmethod
    def load_graph(cls, graph: Union[nx.Graph, nx.DiGraph]) -> "Network":
        """
        Load `Network` object from a `networkx` graph

        Parameters
        ----------
        graph : Union[nx.Graph, nx.DiGraph]
            The networkx graph of the network

        Returns
        -------
        Network
            The instance of the `Network` class
        """
        graph_md = graph.graph
        directed: bool = True if graph_md["directionality"] == "directed" else False
        interaction_type = graph_md["interaction_type"]
        cmetadata = graph_md["computational_metadata"]
        interaction_threshold = cmetadata["interaction_threshold"]
        pvalue_threshold = cmetadata["pvalue_threshold"]
        pvalue_correction = None

        metadata = graph_md
        networkmetadata_model = NetworkmetadataModel(metadata, strict=False)
        networkmetadata_model.validate()

        nodes: List[str] = []
        links: List[LinkDType] = list(graph.edges(data=True))
        lineages: List[dict] = []
        children_map: Dict[str, List[str]] = {}

        for node, ndata in graph.nodes(data=True):
            nodes.append(node)
            lineage = Lineage.from_str(ndata["lineage"]).to_dict("Species")
            children_map[node] = ndata["children"]
            abundance = ndata.get("abundance", np.nan)
            if abundance is not None:
                lineages.append({**lineage, **dict(Abundance=abundance)})
            else:
                lineages.append(lineage)
        obs_metadata = pd.DataFrame(lineages, index=nodes)

        network = cls(
            nodes,
            links,
            metadata,
            cmetadata,
            obs_metadata,
            children_map,
            interaction_type=interaction_type,
            interaction_threshold=interaction_threshold,
            pvalue_threshold=pvalue_threshold,
            pvalue_correction=pvalue_correction,
            directed=directed,
        )
        return network



#!/usr/bin/env python3

from biom import load_table
from Bio import SeqIO


def create_seqtable(seqs, otu_table):
    otu_table_df = otu_table.to_dataframe(dense=True)
    otu_table_df.index = list(map(lambda x: seqs[x], otu_table_df.index))
    otu_table_df.to_csv("seq_table.tsv", sep="\t", index=True)


def main(rep_seqs_file, otu_file):
    seqs = {seq.id: str(seq.seq) for seq in SeqIO.parse(rep_seqs_file, "fasta")}
    otu_table = load_table(otu_file)
    create_seqtable(seqs, otu_table)


if __name__ == "__main__":
    REP_SEQS_FILE = "${repseqs_file}"
    OTU_FILE = "${otutable_file}"
    main(REP_SEQS_FILE, OTU_FILE)



"""
    Module that deals with the validation of an OTU table
"""

import pathlib
from typing import Dict, List, Optional, Union

import pandas as pd
from biom import load_table, Table

from .otu_schema import BiomType, SamplemetaType, ObsmetaType


class OtuValidator:
    """
    Validates input `OTU` table file and returns the `Otu` instance of the file

    Parameters
    ----------
    dtype : {'biom', 'tsv'}
        The type of OtuValidator instance to be created
    ext : str, optional
        The extension of the file if other than supported extensions
        Supported extensions:
        'tsv' dtype: 'tsv', 'txt', 'counts'
        'biom' dtype: 'biom', 'hdf5'

    Attributes
    ----------
    configuration : Dict[str, Any]
        Dictionary showing the current configuration of the instance
    validator : BiomType
        The schmatics validator instance

    Raises
    ------
    ValidationError
        If any of the files do not conform to the schema outlines in `otu_schema`

    Notes
    -----
    We assume that the extension dictates the filetype
    """

    _otu_exts = {"tsv": [".tsv", ".txt", ".counts"], "biom": [".biom", ".hdf5"]}
    _meta_exts = [".csv", ".tsv"]
    _tax_exts = [".csv", ".tsv"]

    def __init__(self, dtype: str, ext: Optional[str] = None) -> None:
        self._dtype = dtype
        if dtype not in self._otu_exts.keys():
            raise TypeError(
                f"{dtype} is not supported. Try one of {self._otu_exts.keys()}"
            )
        if ext:
            self._otu_exts[self._dtype].append(ext)
        self.validator = BiomType()

    @property
    def configuration(self) -> Dict[str, Union[str, List[str]]]:
        """
        Dictionary showing the current configuration of the instance

        Returns
        -------
        Dict[str, Union[str, List[str]]]
        """
        return {
            "dtype": self._dtype,
            "valid_otu_ext": self._otu_exts[self._dtype],
            "valid_meta_ext": self._meta_exts,
            "valid_tax_ext": self._tax_exts,
        }

    def _validate_ext(self, fpath: pathlib.Path) -> bool:
        """
        Determines whether the filetype is supported

        Parameters
        ----------
        fpath : pathlib.Path

        Returns
        -------
        bool
        """
        exts = self._otu_exts[self._dtype]
        return bool(fpath.suffix in exts)

    def _load_from_biom(self, otu_file: pathlib.Path) -> Table:
        """
        Read biom table from file

        Parameters
        ----------
        otu_file : pathlib.Path
            The path to the OTU file in `biom` format

        Returns
        -------
        Table
            A `biom.Table` instance containing the OTU, meta, tax data
        """
        otudata = load_table(otu_file)
        self.validator.validate(otudata)
        return otudata

    @staticmethod
    def _extract_data(data_file: pathlib.Path, valid_exts: List[str]) -> pd.DataFrame:
        """
        Extract data as a `pd.DataFrame` from file

        Parameters
        ----------
        data_file : pathlib.Path
            The path to the data file
        valid_exts : List[str]
            A list of valid extensions

        Returns
        -------
        pd.DataFrame
            `pd.DataFrame` created from the data_file
        """
        ext = data_file.suffix
        if ext in valid_exts:
            if ext == "tsv":
                data = pd.read_table(data_file, sep="\t", index_col=0, na_filter=False)
            elif ext == "csv":
                data = pd.read_csv(data_file, sep=",", index_col=0, na_filter=False)
            else:
                data = pd.read_csv(
                    data_file, sep=None, engine="python", index_col=0, na_filter=False
                )
        else:
            raise TypeError(
                "The input metadata file type is not supported. "
                f"Valid extensions are {valid_exts}"
            )
        return data

    def _load_from_tsv(
        self, otu_file: pathlib.Path, meta_file: pathlib.Path, tax_file: pathlib.Path
    ) -> Table:
        """
        Read OTU counts file to biom table and add metadata and taxonomy data

        Parameters
        ----------
        otu_file : pathlib.Path
            The path to the tsv file containing the OTU counts table
        meta_file : pathlib.Path
            The path to the csv file containing the metadata information
        tax_file : pathlib.Path
            The path to the csv file containing the taxonomy information

        Returns
        -------
        Table
            A `biom.Table` instance containing the OTU, meta, tax data
        """
        otudata = load_table(otu_file)
        metadata = self._extract_data(meta_file, self._meta_exts)
        metadata.index = metadata.index.astype(str)
        samplemeta_type = SamplemetaType()
        samplemeta_type.validate(metadata)
        taxdata = self._extract_data(tax_file, self._tax_exts)
        taxdata.index = taxdata.index.astype(str)
        obsmeta_type = ObsmetaType()
        obsmeta_type.validate(taxdata)
        otudata.add_metadata(metadata.to_dict(orient="index"), axis="sample")
        otudata.add_metadata(taxdata.to_dict(orient="index"), axis="observation")
        self.validator.validate(otudata)
        return otudata

    def load_validate(
        self,
        otu_file: pathlib.Path,
        meta_file: Optional[pathlib.Path] = None,
        tax_file: Optional[pathlib.Path] = None,
    ) -> Table:
        """
        Load the data and validate

        Parameters
        ----------
        otu_file : pathlib.Path
            The path to the `OTU` counts table
        meta_file : pathlib.Path, optional
            The path to the sample metadata file
            This argument is required if `dtype` is 'tsv'
        tax_file : pathlib.Path, optional
            The path to the taxonomy file
            This argument is required if `dtype` is 'tsv'

        Returns
        -------
        Table
            `biom.Table` containing all the data
        """
        err_msg = (
            "The input OTU file type is not supported. "
            f"Valid extensions are {self._otu_exts[self._dtype]}"
        )
        if self._dtype == "biom":
            if self._validate_ext(otu_file):
                otu_table = self._load_from_biom(otu_file)
            else:
                raise ValueError(err_msg)
        elif self._dtype == "tsv":
            if meta_file and tax_file:
                if self._validate_ext(otu_file):
                    otu_table = self._load_from_tsv(otu_file, meta_file, tax_file)
                else:
                    raise TypeError(err_msg)
            else:
                raise ValueError("Missing metadata or taxonomy data")
        return otu_table



#!/usr/bin/env python

from micone import Otu


def main(biom_file, base_name):
    otu_biom = Otu.load_data(biom_file)
    otu_biom.write(base_name=base_name, file_type="tsv")


if __name__ == "__main__":
    BIOM_FILE = "${otu_file}"
    BASE_NAME = "${meta.id}"
    main(BIOM_FILE, BASE_NAME)



#!/usr/bin/env python3

import pathlib
from typing import List

from micone import Network, NetworkGroup


def main(base_name: str, network_files: List[pathlib.Path], id_field: str) -> None:
    networks: List[Network] = []
    for network_file in network_files:
        networks.append(Network.load_json(str(network_file)))
    network_group = NetworkGroup(networks, id_field=id_field)
    pathlib.Path("merged/").mkdir(parents=True, exist_ok=True)
    cids = list(range(len(network_group.contexts)))
    merged_network_group = network_group.combine_pvalues(cids)
    merged_network_group.write(
        "merged/" + base_name + "_network.json", split_files=True
    )


if __name__ == "__main__":
    BASE_NAME = "merged"
    ID_FIELD = "${id_field}"
    NETWORK_FILES = list(pathlib.Path().glob("*_network.json"))
    main(BASE_NAME, NETWORK_FILES, ID_FIELD)
